from __future__ import annotations
import re

import numpy as np
import xarray as xr

from .exceptions import LongitudeError, CoordinateError, AttrError, UnitError


coord_regex = {
    "lon": re.compile("lon[a-z]*"),
    "lat": re.compile("lat[a-z]*"),
    "lev": re.compile("(p?lev|pre?s|isobaric)[a-z]*"),
    "zonal_wavenum": re.compile("(zonal_wavenum|wavenum_lon|lon_wavenum)[a-z]*"),
}

var_units = {
    "altitude": {"m", "meters"},
    "heatflux": {
        "K m s-1",
        "m K s-1",
        "K*m*s-1",
        "m*K*s-1",
        "K * m * s-1",
        "m * K * s-1",
        "K*m/s",
        "m*K/s",
        "K * m / s",
        "m * K / s",
        "Kelvin meters per second",
        "meters Kelvin per second",
    },
    "momentumflux": {
        "m+2 s-2",
        "m+2*s-2",
        "m+2 * s-2",
        "m+2/s+2",
        "m+2 / s+2",
        "meters squared per second squared",
    },
    "prs_vertmomflux": {
        "Pa m s-2",
        "m Pa s-2",
        "Pa*m*s-2",
        "m*Pa*s-2",
        "Pa * m * s-2",
        "m * Pa * s-2",
        "Pa*m/s",
        "m*Pa/s",
        "Pa * m / s",
        "m * Pa / s",
        "Pascal meters per second squared",
        "meter Pascals per second squared",
    },
    "pressure": {"Pa", "Pascals"},
    "temperature": {"K", "degK", "Kelvin"},
    "prs_vvel": {"Pa s-1", "Pa*s-1", "Pa * s-1", "Pa/s", "Pa / s", "Pascals per second"},
    "wind": {"m s-1", "m*s-1", "m * s-1", "m/s", "m / s", "meters per second"},
}


def has_global_regular_lons(
    lons: xr.DataArray | np.ndarray, enforce: bool = False
) -> bool:
    r"""Checks an array of longitudes to ensure it is regularly spaced
    and spans 360 degrees. This is primarily to allow for strict checks
    on data before trying to take zonal means, Fourier transforms, etc.

    Parameters
    ----------
    lons : `xarray.DataArray` or `np.array`
        data containing the longitudes, in degrees

    enforce : bool, optional
        If True, the function will throw an error if the provided
        longitudes are irregular and/or only span a fraction of the globe.
        Defaults to False.

    Returns
    -------
    bool
        True if the lons span 360 degrees and regularly spaced, False if not

    """

    # generally for regularly gridded data, the difference
    # between the max lon and min lon plus the lon delta
    # should equal 360 degrees
    span = float(lons.max() - lons.min())
    dlon = float(lons[1] - lons[0])

    # check that all the longitude differences are close to
    # the single dlon, to within 1.0e-3 degrees
    equal_spacing = np.allclose(np.diff(lons), dlon, rtol=0.0, atol=1.0e-3)

    # now check that the span+londelta cover 360 degrees
    full_span = np.allclose(span + dlon, 360.0, rtol=0.0, atol=1.0e-3)

    if (enforce is True) and (equal_spacing is False):
        msg = "longitudes are not equally spaced"
        raise LongitudeError(msg)
    elif (enforce is True) and (full_span is False):
        msg = "longitudes do not span all 360 degrees"
        raise LongitudeError(msg)

    return equal_spacing and full_span


def infer_xr_coord_names(
    dat: xr.DataArray | xr.Dataset, required: list[str] = []
) -> dict:
    r"""A convenience function that identifies commonly used coordinate names
    for gridded earth datasets. This function enables other functions in
    pyzome to perform operations across coordinates without the user having
    to specify  the coordinate names or change their data to use pre-defined
    coordinate names. Uses regex patterns to identify variations of "lon",
    "lat", and "lev".

    Parameters
    ----------
    dat : `xarray.DataArray` or `xr.Dataset`
        The data containing the coordinates of lat, lon, lev, etc.

    required: List of strings, optional
        If this kwarg is defined, this function will throw errors if it is
        unable to find a name that matches with one of the coordinate-name
        categories defined by coord_regex. Defaults to an empty list,
        meaning the function will not check for any required coordinates.

    Returns
    -------
    coord_names: dict 
        string keys and values, with the keys being the
        coordinate name category (e.g., lat, lon), and the values being the
        actual coordinate name in the given xarray data. E.g,
        {"lat":"latitude"} is a possible return value in which latitude was
        detected as the coordinate for the "lat" category.

    """

    coord_names = {}
    dat_coords = list(dat.coords)
    for dc in dat_coords:
        for coord in coord_regex:
            if coord_regex[coord].match(dc.lower()): # type: ignore
                # Check if more than one coord in dat matches the same pattern
                if coord in coord_names:
                    msg = (
                        "Found multiple coordinates in dat matching the "
                        + f"pattern for '{coord}: {coord_names[coord]} & {dc}"
                    )
                    raise CoordinateError(msg)
                else:
                    coord_names[coord] = dc

    for req in required:
        if req not in coord_names:
            msg = f"Unable to match any of the coordinates in dat for {req}"
            raise CoordinateError(msg)

    return coord_names


def check_var_SI_units(dat: xr.DataArray, var: str, enforce: bool = False) -> bool:
    r"""A function that checks whether the units attribute of a DataArray
    matches SI units for a specific variable category.

    In cases where units matter, pyzome assumes that variables are provided
    in SI units. While there are python libraries that could be used to
    automatically detect and do conversions, pyzome errs on the side of being
    small, simple, and strict (e.g., so that a user would have to explicitly
    do unit conversions themselves, and set units attributes before applying
    a relevant pyzome function).


    Parameters
    ----------
    dat : `xarray.DataArray`
        The data with attributes to check for SI units.

    var: string
        The variable category for checking SI units. These are defined by the
        var_units global dictionary.

    enforce : bool, optional
        If True, the function will throw an error if the provided
        units string in the DataArray does not match the SI units for the
        given variable category. Defaults to False.

    Returns
    -------
    bool
        True if an SI units match is found, False otherwise.

    """

    if "units" not in dat.attrs:
        msg = "units is not an attribute of the given DataArray"
        raise AttrError(msg)

    units_SI = dat.units in var_units[var]
    if (enforce is True) and (units_SI is False):
        msg = (
            f"The units '{dat.units}' do not match SI units for the {var}"
            + " category."
        )
        raise UnitError(msg)

    return units_SI


def check_for_logp_coord(dat: xr.DataArray, enforce: bool = False) -> bool:
    r"""A function that checks whether a log-pressure altitude coordinate
    (assumed to be created by pyzome) exists in the given DataArray. Uses
    a combination of units and long_name to check.

    Parameters
    ----------
    dat : `xarray.DataArray`
        The data to check.

    enforce : bool, optional
        If True, the function will throw an error if the provided
        DataArray does not contain a coordinate named "z" with correct
        units and long_name attributes.

    Returns
    -------
    bool
        True if log-pressure altitude coordinate is found. False otherwise.

    """

    if "z" not in dat.coords:
        if enforce is True:
            msg = "z is not a coordinate in the data"
            raise CoordinateError(msg)
        return False

    if not {"units", "long_name", "note"} <= dat.z.attrs.keys():
        if enforce is True:
            msg = "z is missing at least one of the required attributes 'units', 'long_name', and/or 'note'"
            raise AttrError(msg)
        return False

    if dat.z.attrs["long_name"] != "log-pressure altitude":
        if enforce is True:
            msg = "z must have a long_name = 'log-pressure altitude'"
            raise AttrError(msg)
        return False

    if dat.z.attrs["note"] != "added by pyzome":
        if enforce is True:
            msg = "z must have a note = 'added by pyzome'"
            raise AttrError(msg)
        return False

    return check_var_SI_units(dat.z, "altitude", enforce=enforce)



from __future__ import annotations

import numpy as np
import xarray as xr

from .checks import infer_xr_coord_names, check_var_SI_units
from .constants import (
    PREF,
    GAS_CONST_DRY_AIR,
    SPEC_HEAT_DRY_AIR,
    EARTH_RADIUS,
    EARTH_ROTA_RATE,
)


def resid_vel(
    v: xr.DataArray,
    w: xr.DataArray,
    T: xr.DataArray,
    vT: xr.DataArray,
    p0: float = PREF,
    Rs: float = GAS_CONST_DRY_AIR,
    Cp: float = SPEC_HEAT_DRY_AIR,
    a: float = EARTH_RADIUS,
) -> tuple[xr.DataArray, xr.DataArray]:
    r"""Calculate the residual mean velocity components using zonal mean fields.

    Parameters
    ----------
    v : `xarray.DataArray`
        data containing the zonal mean of the meridional wind
    w : `xarray.DataArray`
        data containing the zonal mean of the vertical *pressure* velocity
    T : `xarray.DataArray`
        data containing the zonal mean of the air temperature
    vT : `xarray.DataArray`
        data containing the zonal mean meridional eddy heat flux, nominally
        defined as zonal_mean((v - zonal_mean(v))*(T - zonal_mean(T)))
    lat_coord : str, optional
        The coordinate name of the latitude dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the latitude
    lev_coord : str, optional
        The coordinate of the pressure-level dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the pressure-levels
    p0 : float, optional
        Reference pressure for computation of potential temperature. Defaults
        to 100000 Pa for Earth.
    Rs : float, optional
        Specific gas constant for computation of potential temperature. Defaults
        to 287.058 J/kg/K for dry air of the Earth.
    Cp : float, optional
        Specific heat capacity at constant pressure for computation of potential
        temperature. Defaults to 1004.64 J/kg/K for dry air of the Earth.
    a : float, optional
        Planetary radius. Defaults to 6.37123e6 m for the Earth.

    Returns
    -------
    residual velocities: tuple of two `xarray.DataArray`
        The meridional and vertical residual velocity components.

    Notes
    -----
    v, w, T, and vT should generally have the same dimensions. However, as
    a consequence of the way xarray performs broadcasting of arrays, this
    function will still work *as long as all the arrays have at least latitude
    and level dimensions*. This has an added benefit that if you desire to
    compute the contribution of different zonal wavenumbers to the residual
    circulation, then you can provide the zonal covariances (vT) with an added
    dimension such as "lon_wavenum" (as is returned by the zonal_wave_covariance
    function) to get the correct result.

    """

    coords = infer_xr_coord_names(v, required=["lat", "lev"])
    lat = coords["lat"]
    lev = coords["lev"]

    check_var_SI_units(v[lev], "pressure", enforce=True)
    check_var_SI_units(v, "wind", enforce=True)
    check_var_SI_units(w, "prs_vvel", enforce=True)
    check_var_SI_units(T, "temperature", enforce=True)
    check_var_SI_units(vT, "heatflux", enforce=True)

    cos_lats = np.cos(np.deg2rad(v[lat]))
    to_theta = (p0 / v[lev]) ** (Rs / Cp)

    Tht = T * to_theta
    vTht = vT * to_theta

    dTht_dp = Tht.differentiate(lev, edge_order=2)
    w_part = (vTht * cos_lats) / dTht_dp
    w_part = (180.0 / np.pi) * w_part.differentiate(lat, edge_order=2)

    v_res = v - (vTht / dTht_dp).differentiate(lev, edge_order=2)
    w_res = w + (1 / (a * cos_lats)) * w_part

    v_res.attrs["units"] = "m s-1"
    w_res.attrs["units"] = "Pa s-1"

    return (v_res, w_res)


def epflux_vector(
    u: xr.DataArray,
    T: xr.DataArray,
    uv: xr.DataArray,
    vT: xr.DataArray,
    uw: xr.DataArray,
    p0: float = PREF,
    Rs: float = GAS_CONST_DRY_AIR,
    Cp: float = SPEC_HEAT_DRY_AIR,
    a: float = EARTH_RADIUS,
    Omega: float = EARTH_ROTA_RATE,
) -> tuple[xr.DataArray, xr.DataArray]:
    r"""Calculate the Eliassen-Palm Flux Vector components using zonal mean fields.

    Parameters
    ----------
    u : `xarray.DataArray`
        data containing the zonal mean of the zonal wind
    T : `xarray.DataArray`
        data containing the zonal mean of the air temperature
    uv : `xarray.DataArray`
        data containing the zonal mean meridional momentum flux
    vT : `xarray.DataArray`
        data containing the zonal mean meridional heat flux
    uw : `xarray.DataArray`
        data containing the zonal mean vertical momentum flux
        (consistent with w, the vertical pressure velocity field)
    p0 : float, optional
        Reference pressure for computation of potential temperature. Defaults
        to 100000 Pa for Earth.
    Rs : float, optional
        Specific gas constant for computation of potential temperature. Defaults
        to 287.058 J/kg/K for dry air of the Earth.
    Cp : float, optional
        Specific heat capacity at constant pressure for computation of potential
        temperature. Defaults to 1004.64 J/kg/K for dry air of the Earth.
    a : float, optional
        Planetary radius. Defaults to 6.37123e6 m for the Earth.
    Omega : float, optional
        Planetary rotation rate. Defaults to 7.29211e-5 s-1 for the Earth.

    Returns
    -------
    ep_flux: tuple of two `xarray.DataArray`
        The meridional and vertical components of the EP-Flux (F_lat, F_prs)

    Notes
    -----
    u, T, uv, vT, and uw should generally have the same dimensions. However,
    as a consequence of the way xarray performs broadcasting of arrays, this
    function will still work *as long as all the arrays have at least latitude
    and level dimensions*. This has an added benefit that if you desire to
    compute EP Fluxes partitioned into contributions from different zonal
    wavenumbers, then you can provide the zonal covariances (uv, vT, and uw)
    with an added dimension such as "zonal_wavenum" (as is returned by the
    zonal_wave_covariance function) to get the correct result.

    """

    coords = infer_xr_coord_names(u, required=["lat", "lev"])
    lat = coords["lat"]
    lev = coords["lev"]

    check_var_SI_units(u[lev], "pressure", enforce=True)
    check_var_SI_units(T, "temperature", enforce=True)
    check_var_SI_units(uv, "momentumflux", enforce=True)
    check_var_SI_units(uw, "prs_vertmomflux", enforce=True)
    check_var_SI_units(vT, "heatflux", enforce=True)

    f = 2 * Omega * np.sin(np.deg2rad(u[lat]))
    cos_lats = np.cos(np.deg2rad(u[lat]))
    to_theta = (p0 / u[lev]) ** (Rs / Cp)

    Tht = T * to_theta
    vTht = vT * to_theta

    dTht_dp = Tht.differentiate(lev, edge_order=2)
    du_dp = u.differentiate(lev, edge_order=2)
    ducos_dphi = (180.0 / np.pi) * (u * cos_lats).differentiate(lat, edge_order=2)

    F_lat = a * cos_lats * ((vTht / dTht_dp) * du_dp - uv)
    F_prs = (
        a
        * cos_lats
        * (
            ((-vTht / dTht_dp) * (1 / (a * cos_lats)) * ducos_dphi)
            + (f * vTht / dTht_dp)
            - uw
        )
    )

    F_lat.attrs["units"] = "m+3 s-2"  # type: ignore
    F_prs.attrs["units"] = "Pa m+2 s-2"  # type: ignore

    return (F_lat, F_prs)  # type: ignore


def qg_epflux_vector(
    T: xr.DataArray,
    uv: xr.DataArray,
    vT: xr.DataArray,
    p0: float = PREF,
    Rs: float = GAS_CONST_DRY_AIR,
    Cp: float = SPEC_HEAT_DRY_AIR,
    a: float = EARTH_RADIUS,
    Omega: float = EARTH_ROTA_RATE,
) -> tuple[xr.DataArray, xr.DataArray]:
    r"""Calculate the quasi-geostrophic Eliassen-Palm Flux Vector
    components using zonal mean fields.

    Parameters
    ----------
    T : `xarray.DataArray`
        data containing the full field air temperature
    uv : `xarray.DataArray`
        data containing the zonal mean meridional momentum flux
    vT : `xarray.DataArray`
        data containing the zonal mean meridional heat flux
    p0 : float, optional
        Reference pressure for computation of potential temperature. Defaults
        to 100000 Pa for Earth.
    Rs : float, optional
        Specific gas constant for computation of potential temperature. Defaults
        to 287.058 J/kg/K for dry air of the Earth.
    Cp : float, optional
        Specific heat capacity at constant pressure for computation of potential
        temperature. Defaults to 1004.64 J/kg/K for dry air of the Earth.
    a : float, optional
        Planetary radius. Defaults to 6.37123e6 m for the Earth.
    Omega : float, optional
        Planetary rotation rate. Defaults to 7.29211e-5 s-1 for the Earth.

    Returns
    -------
    qg_ep_flux: tuple of two `xarray.DataArray`
        The meridional and vertical components of the quasi-geostrophic EP-Flux
        (F_lat, F_prs).

    Notes
    -----
    T, uv, and vT should generally have the same dimensions. However, as a
    consequence of the way xarray performs broadcasting of arrays, this function
    will still work *as long as all the arrays have at least latitude and level
    dimensions*. This has an added benefit that if you desire to compute QG-EP
    Fluxes partitioned into contributions from different zonal wavenumbers, then
    you can provide the zonal covariances (uv and vT) with an added dimension
    such as "zonal_wavenum" (as is returned by the zonal_wave_covariance function)
    to get the correct result.

    """

    coords = infer_xr_coord_names(T, required=["lat", "lev"])
    lat = coords["lat"]
    lev = coords["lev"]

    check_var_SI_units(T[lev], "pressure", enforce=True)
    check_var_SI_units(T, "temperature", enforce=True)
    check_var_SI_units(uv, "momentumflux", enforce=True)
    check_var_SI_units(vT, "heatflux", enforce=True)

    f = 2 * Omega * np.sin(np.deg2rad(T[lat]))
    cos_lats = np.cos(np.deg2rad(T[lat]))
    to_theta = (p0 / T[lev]) ** (Rs / Cp)

    Tht = T * to_theta
    vTht = vT * to_theta

    dTht_dp = Tht.differentiate(lev, edge_order=2)

    F_lat = -a * cos_lats * uv
    F_prs = a * cos_lats * f * (vTht / dTht_dp)

    F_lat.attrs["units"] = "m+3 s-2"  # type: ignore
    F_prs.attrs["units"] = "Pa m+2 s-2"  # type: ignore

    return (F_lat, F_prs)  # type: ignore


def epflux_div(
    F_lat: xr.DataArray,
    F_prs: xr.DataArray,
    accel: bool = False,
    terms: bool = False,
    a: float = EARTH_RADIUS,
) -> xr.DataArray | tuple[xr.DataArray, xr.DataArray]:
    r"""Calculate the Eliassen-Palm Flux divergence assuming
    spherical coordinates.

    Parameters
    ----------
    F_lat : `xarray.DataArray`
        data containing the meridional EP-Flux component
    F_prs : `xarray.DataArray`
        data containing the vertical EP-Flux component
    accel : bool, optional
        If True, will scale the output by 1 / (a*cos(lat)) so
        that the divergence is in units of m s-2
    terms : bool, optional
        If True, the function returns the individual contributions
        to the divergence from the meridional and vertical components
        of the EP-Flux. Defaults to False - the function returns the
        sum of these terms.
    a : float, optional
        Planetary radius. Defaults to 6.37123e6 m for the Earth.

    Returns
    -------
    epflux_divergence: `xarray.DataArray` or tuple of (`xarray.DataArray`, `xarray.DataArray`)
        The total EP-Flux divergence or
        the individual terms from the meridional and vertical divergence (if terms=True)

    See Also
    --------
    epflux_vector
    qg_epflux_vector

    Todo
    -----
    * Add unit checks on input EP-Flux components

    """

    coords = infer_xr_coord_names(F_lat, required=["lat", "lev"])
    lat = coords["lat"]
    lev = coords["lev"]

    check_var_SI_units(F_prs[lev], "pressure", enforce=True)

    cos_lats = np.cos(np.deg2rad(F_lat[lat]))
    scale = 1 / (a * cos_lats)

    merid_div = (
        (180.0 / np.pi) * scale * (F_lat * cos_lats).differentiate(lat, edge_order=2)
    )
    verti_div = F_prs.differentiate(lev, edge_order=2)

    if accel is True:
        merid_div *= scale
        merid_div.attrs["units"] = "m s-2"  # type: ignore
        verti_div *= scale
        verti_div.attrs["units"] = "m s-2"  # type: ignore
    else:
        merid_div.attrs["units"] = "m+2 s-2"  # type: ignore
        verti_div.attrs["units"] = "m+2 s-2"  # type: ignore

    if terms is True:
        return (merid_div, verti_div)  # type: ignore

    with xr.set_options(keep_attrs=True):
        return merid_div + verti_div  # type: ignore



from __future__ import annotations

import numpy as np
import xarray as xr

from .checks import infer_xr_coord_names, check_var_SI_units, check_for_logp_coord
from .constants import (
    PREF,
    RHOREF,
    SCALE_HEIGHT,
    GAS_CONST_DRY_AIR,
    SPEC_HEAT_DRY_AIR,
    EARTH_RADIUS,
    EARTH_ROTA_RATE,
)


def add_logp_altitude(
    dat: xr.Dataset | xr.DataArray,
    lev_coord: str = "",
    H: float = SCALE_HEIGHT,
    p0: float = PREF,
) -> xr.Dataset | xr.DataArray:
    r"""A convenience function for adding a log-pressure altitude coordinate
    to an xarray Dataset or DataArray

    Parameters
    ----------
    dat : `xarray.Dataset` or `xarray.DataArray`
        data containing a pressure coordinate in SI units (Pa)
    lev_coord : string, optional
        The name of the pressure coordinate in the input data. Defaults
        to an empty string, for which the function will try to infer the
        pressure coordinate
    H : float, optional
        Scale height used to compute the log-pressure altitude
    p0 : float, optional
        Reference pressure for computation of log-pressure altitude.
        Defaults to 100000 Pa for Earth.

    Returns
    -------
    `xarray.Dataset` or `xarray.DataArray`
        The input data with an added log-pressure altitude coordinate named 'z'

    """

    if lev_coord == "":
        coords = infer_xr_coord_names(dat, required=["lev"])
        lev_coord = coords["lev"]
    check_var_SI_units(dat[lev_coord], "pressure", enforce=True)

    z = -H * np.log(dat[lev_coord] / p0)
    z.attrs["units"] = "m"  # type: ignore
    z.attrs["long_name"] = "log-pressure altitude"  # type: ignore
    z.attrs["note"] = "added by pyzome"  # type: ignore

    return dat.assign_coords({"z": z})


def buoyancy_frequency_squared(
    T: xr.DataArray,
    Rs: float = GAS_CONST_DRY_AIR,
    Cp: float = SPEC_HEAT_DRY_AIR,
    H: float = SCALE_HEIGHT,
    p0: float = PREF,
) -> xr.DataArray:
    r"""Calculates the buoyancy frequency squared given temperature.

    Parameters
    ----------
    T : `xarray.DataArray`
        The temperature data in units of Kelvin
    Rs : float, optional
        Specific gas constant. Defaults to 287.058 J/kg/K for dry
        air of the Earth.
    Cp : float, optional
        Specific heat capacity at constant pressure. Defaults to
        1004.64 J/kg/K for dry air of the Earth.
    H : float, optional
        The scale height used to calculate the log-pressure altitude.
        Defaults to 7000 m.
    p0 : float, optional
        Reference pressure used to calculate the log-pressure altitude.
        Defaults to 100000 Pa for Earth.

    Returns
    -------
    Nsq: `xarray.DataArray`
        The buoyancy frequency squared, in units of s-2.

    """

    check_for_logp_coord(T, enforce=True)
    check_var_SI_units(T, "temperature", enforce=True)

    dT_dz = T.differentiate("z", edge_order=2)
    Nsq = (Rs / H) * (dT_dz + (Rs / Cp) * (T / H))
    Nsq.attrs["units"] = "s-2"
    Nsq.attrs["long_name"] = "buoyancy frequency squared"

    return Nsq


def merid_grad_qgpv(
    u: xr.DataArray,
    Nsq: xr.DataArray | float,
    lat_coord: str = "",
    rho_s: float = RHOREF,
    H: float = SCALE_HEIGHT,
    Omega: float = EARTH_ROTA_RATE,
    a: float = EARTH_RADIUS,
    terms: bool = False,
) -> xr.DataArray | tuple[xr.DataArray, xr.DataArray, xr.DataArray]:
    r"""Calculates the meridional gradient in the quasi-geostrophic
    potential vorticity (QGPV) given the zonal mean zonal winds, and
    the squared buoyancy frequency.

    Parameters
    ----------
    u : `xarray.DataArray`
        The zonal mean zonal wind data.
    Nsq : `xarray.DataArray` or float
        The squared buoyancy frequency. Nsq need not have the same dimensions
        as u, but it should be consistent with u in the sense that it must be
        able to be properly broadcasted when used in computations with u (it is
        common to use a reference Nsq that only varies with altitude, or
        latitude/altitude)
    lat_coord : str, optional
        The coordinate name of the latitude dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the latitude
    rho_s : float, optional
        The reference density for air at the surface. Defaults to 1.2 kg m-3
        for density of air on Earth.
    H : float, optional
        The scale height used to calculate the log-pressure altitude.
        Defaults to 7000 m.
    Omega : float, optional
        Planetary rotation rate. Defaults to 7.29211e-5 s-1 for the Earth.
    a : float, optional
        Planetary radius. Defaults to 6.37123e6 m for the Earth.
    terms : bool, optional
        If False (the default), the function returns the sum of the 3 terms
        making up the QGPV. If True, the function returns the 3 individual
        terms as a tuple containing the contributions from (1) the change in
        planetary vorticity with latitude; (2) the horizontal curvature of the
        zonal mean zonal wind; and (3) the vertical curvature of the zonal mean
        zonal wind.

    Returns
    -------
    qgpv_grad: `xarray.DataArray` or tuple of DataArrays
        Depending on the value of the terms keyword argument, either the full
        meridional QGPV gradient field, or a tuple of the individual terms
        that make up the total, in units of s-1.

    Todo
    -----
    * Unit checks on u and Nsq (if Nsq is a DataArray)
    * Add units on output terms

    """

    check_for_logp_coord(u, enforce=True)
    # note that Nsq is *not* checked like u, since a user may want to
    # use a reference Nsq independent of latitude and/or time

    if lat_coord == "":
        coords = infer_xr_coord_names(u, required=["lat"])
        lat_coord = coords["lat"]

    r2d = 180.0 / np.pi
    phi = np.deg2rad(u[lat_coord])
    cosphi = np.cos(phi)
    f = 2 * Omega * np.sin(phi)

    # contribution from change in planetary vorticity
    grad_coriolis = 2 * Omega * cosphi

    # contribution from meridional wind curvature
    horiz_curv = (
        r2d * (u * cosphi).differentiate(lat_coord, edge_order=2) / (a * cosphi)
    )
    horiz_curv = -r2d * horiz_curv.differentiate(lat_coord, edge_order=2)

    # contribution from vertical wind curvature
    rho_0 = rho_s * np.exp(-u.z / H)
    du_dz = u.differentiate("z", edge_order=2)
    verti_curv = (-a * f * f / rho_0) * ((rho_0 / Nsq) * du_dz).differentiate(
        "z", edge_order=2
    )

    # return individual terms if requested, otherwise the sum
    if terms is True:
        return (grad_coriolis, horiz_curv, verti_curv)  # type: ignore
    else:
        return grad_coriolis + horiz_curv + verti_curv


def refractive_index(
    u: xr.DataArray,
    q_phi: xr.DataArray,
    Nsq: xr.DataArray | float,
    k: int,
    phase_speed: float = 0,
    lat_coord: str = "",
    rho_s: float = RHOREF,
    H: float = SCALE_HEIGHT,
    Omega: float = EARTH_ROTA_RATE,
    a: float = EARTH_RADIUS,
    cd_approx_term: bool = False,
    terms: bool = False,
) -> xr.DataArray | tuple[xr.DataArray, xr.DataArray, xr.DataArray]:
    r"""Calculates the (squared) refractive index for a given distribution of
    zonal mean zonal winds, meridional QGPV gradients, buoyancy frequency
    squared, and zonal wavenumber+phase speed.

    Parameters
    ----------
    u : `xarray.DataArray`
        The zonal mean zonal wind data.
    q_phi : `xarray.DataArray`
        The meridional QGPV gradient
    Nsq : `xarray.DataArray` or float
        The squared buoyancy frequency. Nsq need not have the same dimensions
        as u, but it should be consistent with u and q_phi in the sense that it
        must be able to be properly broadcasted when used in computations (it is
        common to use a reference Nsq that is constant value or time-mean field)
    k : int
        The zonal wavenumber
    phase_speed : float, optional
        The phase speed of the wave, in m s-1.
    lat_coord : str, optional
        The coordinate name of the latitude dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the latitude
    rho_s : float, optional
        The reference density for air at the surface. Defaults to 1.2 kg m-3
        for density of air on Earth.
    H : float, optional
        The scale height used to calculate the log-pressure altitude.
        Defaults to 7000 m.
    Omega : float, optional
        Planetary rotation rate. Defaults to 7.29211e-5 s-1 for the Earth.
    a : float, optional
        Planetary radius. Defaults to 6.37123e6 m for the Earth.
    cd_approx_term : bool, optional
        If False (the default), the function uses the standard buoyancy
        frequency term involving only Nsq. If True, the function uses an
        approximation originally derived by Charney & Drazin involving
        the vertical curvature of the buoyancy frequency. See Weinberger
        et al., 2021, "The Efficiency of Upward Wave Propagation near the
        Tropopause: Importance of the Form of the Refractive Index" for
        details of this term.
    terms : bool, optional
        If False (the default), the function returns the sum of the 3 terms
        making up the squared refractive index. If True, the function returns
        the 3 individual terms as a tuple containing the contributions from
        (1) the contribution from the meridional QGPV gradient; (2) the zonal
        wavenumber; and (3) the buoyancy.

    Returns
    -------
    RIsq: `xarray.DataArray` or tuple of DataArrays
        Depending on the value of the terms keyword argument, either the full
        squared refractive index field, or a tuple of the individual terms
        that make the total, in units of m-2.

    Todo
    -----
    * Unit checks on u, Nsq (if Nsq is a DataArray), and q_phi
    * Add units on output terms

    """

    check_for_logp_coord(u, enforce=True)
    check_for_logp_coord(q_phi, enforce=True)
    if lat_coord == "":
        coords = infer_xr_coord_names(u, required=["lat"])
        lat_coord = coords["lat"]

    lats = np.deg2rad(u[lat_coord])
    cosphi = np.cos(lats)
    f = 2 * Omega * np.sin(lats)

    qgpv_grad_term = q_phi / (a * (u - phase_speed))
    wavenum_term = -k * k / (a * a * cosphi * cosphi)

    # See Weinberger et al., 2021 for reference
    if cd_approx_term is True:
        rho_0 = rho_s * np.exp(-u.z / H)
        d_dz = np.sqrt(rho_0 / Nsq).differentiate("z", edge_order=2)
        d2_dz2 = d_dz.differentiate("z", edge_order=2)
        buoyancy_term = -(f * f / np.sqrt(Nsq)) * (1 / np.sqrt(rho_0)) * d2_dz2
    else:
        buoyancy_term = -f * f / (4 * Nsq * H * H)

    if terms is True:
        return (qgpv_grad_term, wavenum_term, buoyancy_term)  # type: ignore
    else:
        return qgpv_grad_term + wavenum_term + buoyancy_term


def plumb_wave_activity_flux(
    psip: xr.DataArray,
    Nsq: xr.DataArray | float,
    components: list[str] = ["x", "y", "z"],
    lat_coord: str = "",
    lon_coord: str = "",
    Omega: float = EARTH_ROTA_RATE,
    a: float = EARTH_RADIUS,
) -> list[xr.DataArray]:
    r"""Calculates the components of the Plumb wave activity flux given the eddy
    streamfunction and buoyancy frequency squared.

    Parameters
    ----------
    psip : `xarray.DataArray`
        The eddy streamfunction in units of m+2 s-1
    Nsq : `xarray.DataArray` or float
        The squared buoyancy frequency. Nsq need not have the same dimensions
        as psip, but it should be consistent with psip in the sense that it
        must be able to be properly broadcasted when used in computations (it is
        common to use a reference Nsq that is constant value or time-mean field)
    components : list, optional
        The components of the wave activity flux to compute and return in the
        output. 'x' refers to the east-west component, 'y' refers to the north-
        south component, and 'z' refers to the vertical component.
    lat_coord : str, optional
        The coordinate name of the latitude dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the latitude
    lon_coord : str, optional
        The coordinate name of the longitude dimension. If given an empty string
        (the default), the function tries to infer which coordinate corresponds
        to the longitude
    Omega : float, optional
        Planetary rotation rate. Defaults to 7.29211e-5 s-1 for the Earth.
    a : float, optional
        Planetary radius. Defaults to 6.37123e6 m for the Earth.

    Returns
    -------
    waf: list of `xarray.DataArray`s
        The wave activity flux vector components consistent with the
        desired components keyword argument.

    Todo
    -----
    * Unit checks on psip and Nsq (if Nsq is a DataArray)
    * Add units on output terms

    """

    check_for_logp_coord(psip, enforce=True)
    coords = infer_xr_coord_names(psip, required=["lev"])

    if ("y" in components) and ("lat" not in coords):
        msg = "A latitude coordinate must be available"
        raise ValueError(msg)
    elif (("x" in components) or ("z" in components)) and ("lon" not in coords):
        msg = "A longitude coordinate must be available"
        raise ValueError(msg)

    lat_coord = coords["lat"]
    lon_coord = coords["lon"]
    r2d = 180.0 / np.pi
    lats = np.deg2rad(psip[lat_coord])
    cosphi = np.cos(lats)
    f = 2 * Omega * np.sin(lats)

    p = psip[coords["lev"]] / 100000.0

    waf = []
    dpsi_dlam = r2d * psip.differentiate(lon_coord, edge_order=2)
    if "x" in components:
        d2psi_dlam2 = r2d * dpsi_dlam.differentiate(lon_coord, edge_order=2)
        wafx = (p / (2 * a * a * cosphi)) * ((dpsi_dlam) ** 2 - psip * d2psi_dlam2)
        waf.append(wafx)
    if "y" in components:
        dpsi_dphi = r2d * psip.differentiate(lat_coord, edge_order=2)
        d2psi_dphidlam = r2d * dpsi_dphi.differentiate(lon_coord, edge_order=2)
        wafy = (p / (2 * a * a)) * (dpsi_dlam * dpsi_dphi - psip * d2psi_dphidlam)
        waf.append(wafy)
    if "z" in components:
        dpsi_dz = psip.differentiate("z", edge_order=2)
        d2psi_dlamdz = dpsi_dlam.differentiate("z", edge_order=2)
        wafz = (p * f * f / (2 * Nsq * a)) * (dpsi_dlam * dpsi_dz - psip * d2psi_dlamdz)
        waf.append(wafz)

    return waf



import pytest

import numpy as np


from pyzome.mock_data import (
    create_dummy_geo_field,
    create_dummy_geo_dataset,
    _make_regular_coord,
    lon_coord,
    lat_coord,
    plev_coord,
    time_coord,
)


def test_reg_coord_start_and_step():
    """Test that regular coordinate is created correctly"""
    da = _make_regular_coord(1, 10, 2, "coord", False)
    np.testing.assert_allclose(da.values, np.arange(1, 10, 2))

    da = _make_regular_coord(0, 10, 2, "coord", False)
    np.testing.assert_allclose(da.values, np.arange(0, 10, 2))


def test_reg_coord_inclusive():
    """Test that _make_regular_coord is inclusive of the right-side value"""
    da = _make_regular_coord(0, 10, 2, "coord", True)
    np.testing.assert_allclose(da.values, np.arange(0, 12, 2))

    da = _make_regular_coord(0, 10, 2, "coord", False)
    np.testing.assert_allclose(da.values, np.arange(0, 10, 2))


def test_reg_coord_custom_name():
    """Test that _make_regular_coord can set a custom name"""
    da = _make_regular_coord(1, 10, 2, "testname", False)
    assert da.name == "testname"


def test_reg_coord_invalid_resolution():
    """
    Test that _make_regular_coord raises ValueError for invalid step size
    for the given limits
    """
    with pytest.raises(ValueError) as e:
        _make_regular_coord(90, -90, 10, "coord", True)
    assert "Invalid step for the given limits" in str(e.value)

    with pytest.raises(ValueError):
        _make_regular_coord(-180, 180, -10, "coord", False)
    assert "Invalid step for the given limits" in str(e.value)


def test_reg_coord_zero_step():
    """Test that _make_regular_coord raises ValueError for zero step size"""
    with pytest.raises(ValueError) as e:
        _make_regular_coord(-180, 180, 0, "coord", False)
    assert "step cannot be zero" in str(e.value)


def test_reg_coord_equal_step():
    """Test that _make_regular_coord raises ValueError for equal left and right values"""
    with pytest.raises(ValueError) as e:
        _make_regular_coord(10, 10, 1, "coord", False)
    assert "left- and right-side values cannot be equal" in str(e.value)


def test_lats_default_values():
    """Test that lat_coord creates a latitude-like coordinate correctly for default values"""
    da = lat_coord(30)
    assert da.name == "lat"
    np.testing.assert_allclose(da.values, np.arange(-90, 91, 30))
    assert da.attrs["standard_name"] == "latitude"
    assert da.attrs["units"] == "degrees_north"


def test_lats_custom_values():
    """Test that lat_coord creates a latitude-like coordinate correctly for custom values"""
    da = lat_coord(20, "latitude", -80, 80, False)
    assert da.name == "latitude"
    np.testing.assert_allclose(da.values, np.arange(-80, 80, 20))

    da = lat_coord(-20, "latitude", 80, -80, False)
    np.testing.assert_allclose(da.values, np.arange(80, -80, -20))


def test_lats_invalid_latitudes():
    """Test that lat_coord raises ValueError for invalid latitude limits"""
    with pytest.raises(ValueError) as e:
        lat_coord(10, left_lim=-91, right_lim=90)
    assert "Latitudes cannot be below -90 or above 90" in str(e.value)

    with pytest.raises(ValueError) as e:
        lat_coord(10, left_lim=-90, right_lim=91)
    assert "Latitudes cannot be below -90 or above 90" in str(e.value)


def test_lons_default_values():
    """Test that lon_coord creates a longitude-like coordinate correctly for default values"""
    da = lon_coord(30)
    assert da.name == "lon"
    np.testing.assert_allclose(da.values, np.arange(0, 360, 30))
    assert da.attrs["standard_name"] == "longitude"
    assert da.attrs["units"] == "degrees_east"


def test_lons_custom_values():
    """Test that lon_coord creates a longitude-like coordinate correctly for custom values"""
    da = lon_coord(45, "longitude", 0, 180, True)
    assert da.name == "longitude"
    np.testing.assert_allclose(da.values, np.arange(0, 181, 45))


def test_invalid_longitude_limits():
    """Test that lon_coord raises ValueError for invalid longitude limits"""
    with pytest.raises(ValueError) as e:
        lon_coord(30, "longitude", -180, 400)
    assert "Proper longitude range should be within [-180,180] or [0, 360]" in str(
        e.value
    )

    with pytest.raises(ValueError) as e:
        lon_coord(30, "longitude", 200, -10)
    assert "Proper longitude range should be within [-180,180] or [0, 360]" in str(
        e.value
    )

    with pytest.raises(ValueError) as e:
        lon_coord(30, "longitude", -200, 180)
    assert "lower longitude limit cannot be below -180" in str(e.value)

    with pytest.raises(ValueError) as e:
        lon_coord(30, "longitude", 30, 390)
    assert "upper longitude limit cannot exceed 360" in str(e.value)


def test_plev_default_values():
    """Test that plev_coord creates a pressure-like coordinate correctly for default values"""
    da = plev_coord(3)
    assert da.name == "lev"
    np.testing.assert_allclose(da.values, np.logspace(3, 0, 3 * 3 + 1))
    assert da.attrs["standard_name"] == "pressure"
    assert da.attrs["units"] == "hPa"


def test_plev_custom_values():
    """Test that plev_coord creates a pressure-like coordinate correctly for custom values"""
    da = plev_coord(5, 2, 0, "pressure", "Pa")
    assert da.name == "pressure"
    np.testing.assert_allclose(da.values, np.logspace(2, 0, 5 * 2 + 1))
    assert da.attrs["standard_name"] == "pressure"
    assert da.attrs["units"] == "Pa"


def test_plev_negative_levels_per_decade():
    """Test that plev_coord raises ValueError for negative levels_per_decade"""
    with pytest.raises(ValueError) as e:
        plev_coord(-3)
    assert "'levels_per_decade' must be greater than 0" in str(e.value)


def test_plev_same_exponent_limits():
    """Test that plev_coord raises ValueError for same left_lim_exponent and right_lim_exponent"""
    with pytest.raises(ValueError) as e:
        plev_coord(3, 2, 2)
    assert "'left_lim_exponent' and 'right_lim_exponent' must not be the same" in str(
        e.value
    )


def test_time_default_values():
    """Test that time_coord creates a time-like coordinate correctly for default values"""
    da = time_coord()
    assert da.name == "time"
    np.testing.assert_array_equal(
        da.values,
        np.arange("2000-01-01", "2001-01-01", dtype="datetime64[M]").astype(
            "datetime64[ns]"
        ),
    )
    assert da.attrs["standard_name"] == "time"
    assert da.attrs["units"] == "days since 2000-01-01"


def test_time_custom_values():
    """Test that time_coord creates a time-like coordinate correctly for custom values"""
    da = time_coord("2000-01-01", "2002-01-01", "D")
    assert da.name == "time"
    np.testing.assert_array_equal(
        da.values,
        np.arange("2000-01-01", "2002-01-01", dtype="datetime64[D]").astype(
            "datetime64[ns]"
        ),
    )


def test_time_start_after_end():
    """Test that time_coord raises ValueError for start time after end time"""
    with pytest.raises(ValueError) as e:
        time_coord("2001-01-02", "2000-01-01")
    assert "start time must be before end time" in str(e.value)


def test_invalid_frequency():
    """Test that time_coord raises a TypeError for invalid frequency"""
    with pytest.raises(TypeError):
        time_coord("2000-01-01", "2002-01-01", "Z")


def test_create_dummy_geo_field():
    """Test that create_dummy_geo_field creates a dummy field correctly"""
    lons = lon_coord(10)
    lats = lat_coord(10)
    levs = plev_coord(3)
    times = time_coord()

    da = create_dummy_geo_field(
        lons, lats, levs, times, name="test_field", attrs={"unit": "K"}
    )

    # Check the dimensions
    assert da.dims == ("time", "lev", "lat", "lon")

    # Check the sizes
    assert da.sizes == {
        "time": times.size,
        "lev": levs.size,
        "lat": lats.size,
        "lon": lons.size,
    }

    # Check the name
    assert da.name == "test_field"

    # Check the attributes
    assert da.attrs == {"unit": "K"}

    # Check without levels and time
    da = create_dummy_geo_field(lons, lats)

    # Check the dimensions
    assert da.dims == ("lat", "lon")

    # Check the sizes
    assert da.sizes == {"lat": lats.size, "lon": lons.size}


def test_create_dummy_geo_field_no_dimensions():
    """Test that create_dummy_geo_field raises a ValueError for no dimensions"""
    with pytest.raises(ValueError) as e:
        create_dummy_geo_field(None, None, None, None)
    assert "At least one of 'lons', 'lats', 'levs', or 'times' must be provided" in str(
        e.value
    )


def test_create_dummy_geo_field_invalid_arguments():
    """Test that create_dummy_geo_field raises errors for invalid arguments"""
    invalid_dataarray = "not a DataArray"
    lons = lon_coord(10)
    lats = lat_coord(10)

    with pytest.raises(AttributeError):
        create_dummy_geo_field(invalid_dataarray, lats)

    with pytest.raises(AttributeError):
        create_dummy_geo_field(lons, invalid_dataarray)


def test_create_dummy_geo_dataset():
    """Test that create_dummy_geo_dataset creates a dummy Dataset correctly"""
    lons = lon_coord(10)
    lats = lat_coord(10)
    levs = plev_coord(5)
    times = time_coord()
    field_names = ["temp", "precip", "wind"]
    field_attrs = {
        "temp": {"units": "K"},
        "precip": {"units": "mm/hr"},
        "wind": {"units": "m/s"},
    }

    ds = create_dummy_geo_dataset(field_names, lons, lats, levs, times, field_attrs)

    # Check if the created Dataset has the expected DataArrays
    for name in field_names:
        assert name in ds

        # Check if the DataArray has the expected attributes
        assert ds[name].attrs == field_attrs[name]

        # Check the dimensions
        assert ds[name].dims == ("time", "lev", "lat", "lon")

        # Check the sizes
        assert ds[name].sizes == {
            "time": times.size,
            "lev": levs.size,
            "lat": lats.size,
            "lon": lons.size,
        }


def test_create_dummy_geo_dataset_invalid_arguments():
    """Test that create_dummy_geo_dataset raises errors for invalid arguments"""
    invalid_dataarray = "not a DataArray"
    lons = lon_coord(10)
    lats = lat_coord(10)
    field_names = ["temp"]

    with pytest.raises(AttributeError):
        create_dummy_geo_dataset(field_names, invalid_dataarray, lats)

    with pytest.raises(AttributeError):
        create_dummy_geo_dataset(field_names, lons, invalid_dataarray)

    mismatched_field_attrs = {"temp": {"units": "K"}, "missing_field": {}}
    with pytest.raises(ValueError) as e:
        create_dummy_geo_dataset(
            field_names, lons, lats, field_attrs=mismatched_field_attrs
        )
    assert (
        "Number of field names (1) must match number of keys in field attributes (2)"
        in str(e.value)
    )



PREF = 100000.0  # Reference pressure for surface: Pa
RHOREF = 1.2  # Reference density for surface: kg / m+3
SCALE_HEIGHT = 7000  # Reference Scale-height: m

GAS_CONST_DRY_AIR = 287.058  # Gas constant for dry air: J / K / kg
SPEC_HEAT_DRY_AIR = 1004.64  # Spec. heat for dry air (const pressure): J / K / kg

EARTH_RADIUS = 6.37123e6  # Radius of Earth: m
EARTH_ROTA_RATE = 7.29211e-5  # Rotation rate of Earth: s-1



from __future__ import annotations

import warnings
from typing import Iterable

import numpy as np
import xarray as xr
import scipy
import xrft

from .checks import has_global_regular_lons, infer_xr_coord_names
from .mock_data import lon_coord


def zonal_wave_coeffs(
    dat: xr.DataArray,
    *,
    waves: None | Iterable[int] = None,
    fftpkg: str = "scipy",
    lon_coord: str = "",
) -> xr.DataArray:
    r"""Compute the Fourier transform of data along the zonal direction.

    This function is imported at the top level of the package by default.

    Parameters
    ----------
    dat : `xarray.DataArray`
        input data containing a longitude dimension that spans all 360 degrees
    waves : array-like, optional
        The zonal wavenumbers to maintain in the output. Defaults to None
        for all.
    fftpkg : string, optional
        String that specifies how to perform the FFT on the data. Options are
        scipy or xrft. Specifying scipy uses some operations that are
        memory-eager and leverages scipy.fft.rfft. Specifying xrft should
        leverage the benefits of xarray/dask for large datasets by using
        xrft.fft. Defaults to scipy.
    lon_coord : str, optional
        The coordinate name of the longitude dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the longitude

    Returns
    -------
    `xarray.DataArray`
        Output of an rFFT along the longitude dimension, for specified
        waves only. The output is a complex-valued DataArray containing
        at least the `zonal_wavenum` dimension, with attributes `nlons`
        and `lon0` that specify the number of longitudes and the starting
        longitude of the input data.

    """

    if fftpkg not in ["scipy", "xrft"]:
        msg = "fftpkg must be 'scipy' or 'xrft'"
        raise ValueError(msg)

    if lon_coord == "":
        coords = infer_xr_coord_names(dat, required=["lon"])
        lon_coord = coords["lon"]
    has_global_regular_lons(dat[lon_coord], enforce=True)

    funcs = {"scipy": _zonal_wave_coeffs_scipy, "xrft": _zonal_wave_coeffs_xrft}
    fc = funcs[fftpkg](dat, lon_coord)

    fc.attrs["nlons"] = dat[lon_coord].size
    fc.attrs["lon0"] = dat[lon_coord].values[0]
    if waves is not None:
        fc = fc.sel(zonal_wavenum=waves)

    return fc


def _zonal_wave_coeffs_scipy(dat: xr.DataArray, lon_coord: str) -> xr.DataArray:
    r"""Calculate the Fourier coefficients of waves in the zonal direction.

    Uses scipy.fft.rfft to perform the calculation.

    Parameters
    ----------
    dat : `xarray.DataArray`
        data containing a dimension named longitude that spans all 360 degrees
    lon_coord : string
        name of the dimension/coordinate corresponding to the longitudes

    Returns
    -------
    `xarray.DataArray`
        Output of the rFFT along the longitude dimension.

    """
    nlons = dat[lon_coord].size
    lon_ax = dat.get_axis_num(lon_coord)

    new_dims = list(dat.dims)
    new_dims[lon_ax] = "zonal_wavenum"  # type: ignore

    new_coords = dict(dat.coords)
    new_coords.pop(lon_coord)
    new_coords["zonal_wavenum"] = np.arange(0, nlons // 2 + 1)

    fc = scipy.fft.rfft(dat.values, axis=lon_ax)
    fc = xr.DataArray(fc, coords=new_coords, dims=new_dims)

    return fc


def _zonal_wave_coeffs_xrft(dat: xr.DataArray, lon_coord: str) -> xr.DataArray:
    r"""Calculate the Fourier coefficients of waves in the zonal direction.

    Uses xrft.fft to perform the calculation.

    Parameters
    ----------
    dat : `xarray.DataArray`
        data containing a dimension named longitude that spans all 360 degrees
    lon_coord : string
        name of the dimension/coordinate corresponding to the longitudes

    Returns
    -------
    `xarray.DataArray`
        Output of the rFFT along the longitude dimension.

    """

    fc = xrft.fft(
        dat, dim=lon_coord, real_dim=lon_coord, true_phase=False, true_amplitude=False
    )

    fc = fc.rename({f"freq_{lon_coord}": "zonal_wavenum"})
    fc = fc.assign_coords({"zonal_wavenum": np.arange(fc.zonal_wavenum.size)})

    return fc


def inflate_zonal_wave_coeffs(
    fc_subset: xr.DataArray,
    wave_coord: str = "",
) -> xr.DataArray:
    r"""Inflates an array of zonal wavenumber Fourier coefficients to
    its expected full-spectrum size. The full size is determined by
    the "nlons" attribute of the input DataArray.

    Inflating is done by inserting zeros in the missing wavenumber bins.
    This can be useful for keeping only a subset of wavenumbers while still
    allowing for taking an inverse FFT for filtering or determining the
    contributions of individual wavenumbers to the full field.

    Parameters
    ----------
    fc_subset : `xarray.DataArray`
        Fourier coefficients as a function of zonal wavenumber, as
        returned by `zonal_wave_coeffs`.
    wave_coord : str, optional
        The coordinate name of the wavenumber dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the wavenumbers

    Returns
    -------
    `xarray.DataArray`
        Fourier coefficients as a function of zonal wavenumber, with the
        spectrum size determined by the "nlons" attribute of the input
        DataArray.

    """
    if wave_coord == "":
        coords = infer_xr_coord_names(fc_subset, required=["zonal_wavenum"])
        wave_coord = coords["zonal_wavenum"]

    if "nlons" not in fc_subset.attrs:
        msg = "input DataArray must have an 'nlons' attribute specifying the number of longitudes in the source data"
        raise KeyError(msg)

    expected_wavenums = np.arange(fc_subset.attrs["nlons"] // 2 + 1)
    if np.array_equiv(fc_subset[wave_coord], expected_wavenums):
        return fc_subset
    elif np.any(~np.in1d(fc_subset[wave_coord], expected_wavenums, assume_unique=True)):
        msg = "input DataArray wavenumbers are not a subset of the expected wavenumbers based on the 'nlons' attribute"
        raise ValueError(msg)

    return fc_subset.reindex({wave_coord: expected_wavenums}, fill_value=0j)  # type: ignore


def zonal_wave_ampl_phase(
    fc: xr.DataArray,
    phase_deg: bool = False,
    wave_coord: str = "",
) -> tuple[xr.DataArray, xr.DataArray]:
    r"""Calculates the amplitudes and relative phases of waves in the
    zonal direction.

    Parameters
    ----------
    fc : `xarray.DataArray`
        Fourier coefficients as a function of zonal wavenumber, as
        returned by `zonal_wave_coeffs`.
    phase_deg : boolean, optional
        Whether to return the relative phases in radians or degrees.
    wave_coord : str, optional
        The coordinate name of the wavenumber dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the wavenumbers

    Returns
    -------
    tuple of two `xarray.DataArray`
        Contains the amplitudes and phases for the input field zonal waves.

    See Also
    --------
    zonal_wave_coeffs

    """

    if wave_coord == "":
        coords = infer_xr_coord_names(fc, required=["zonal_wavenum"])
        wave_coord = coords["zonal_wavenum"]

    if "nlons" not in fc.attrs:
        msg = "input DataArray must have an 'nlons' attribute specifying the number of longitudes in the source data"
        raise KeyError(msg)

    # where the longitudinal wavenumber is 0, `where' will
    # mask to nan, so np.isfinite will return False in those
    # spots and true everywhere else. Thus, add 1 to get
    # the multiplying mask that keeps in mind the "zeroth"
    # mode (which would be the zonal mean, if kept)
    #
    # this is necessary because of the symmetric spectrum,
    # so all other wavenumbers except the 0th need to
    # be multipled by 2 to get the right amplitude
    mult_mask = np.isfinite(fc.where(fc[wave_coord] != 0)) + 1

    with xr.set_options(keep_attrs=True):
        ampl = np.absolute(fc) * mult_mask / fc.nlons
        # phas = np.angle(fc, deg=phase_deg) # returns a np.ndarray instead of xr.DataArray/Dataset
        phas = xr.apply_ufunc(np.angle, fc, kwargs={"deg": phase_deg})

    return (ampl.astype(fc.dtype), phas.astype(fc.dtype))


def filter_by_zonal_wave_truncation(
    fc: xr.DataArray,
    waves: Iterable[int],
    fftpkg: str = "scipy",
    wave_coord: str = "",
    lons: None | xr.DataArray = None,
) -> xr.DataArray:
    """Filters a field by truncating the zonal wavenumbers. This is done
    by taking an inverse rFFT of the Fourier coefficients, with the unwanted
    wavenumbers set to zero.

    Parameters
    ----------
    fc : `xarray.DataArray`
        Fourier coefficients as a function of zonal wavenumber, as
        returned by `zonal_wave_coeffs`.
    waves : iterable of int or slice
        The wavenumbers to keep. If a slice is given, it must be a slice
        of integers. If an iterable is given, it must be an iterable of
        integers.
    fftpkg : {'scipy', 'numpy'}, optional
        The FFT package to use. Defaults to 'scipy'.
    wave_coord : str, optional
        The coordinate name of the wavenumber dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the wavenumbers
    lons: `xarray.DataArray`, optional
        The longitude coordinate of the input field. If not given, the
        function will attempt to reconstruct the coordinate from the input
        Fourier coefficients, by looking for 'nlons' and 'lon0' attributes.

    """
    if wave_coord == "":
        coords = infer_xr_coord_names(fc, required=["zonal_wavenum"])
        wave_coord = coords["zonal_wavenum"]

    if "nlons" not in fc.attrs:
        msg = "input DataArray must have an 'nlons' attribute specifying the number of longitudes in the source data"
        raise KeyError(msg)

    if "lon0" not in fc.attrs:
        msg = "input DataArray must have a 'lon0' attribute specifying the starting longitude of the source data"
        raise KeyError(msg)

    wavenum_ax = fc.get_axis_num(wave_coord)
    if lons is None:
        warnings.warn("attempting to infer the input longitude coordinate", UserWarning)
        lons = lon_coord(
            360 / fc.attrs["nlons"],
            left_lim=fc.attrs["lon0"],
            right_lim=fc.attrs["lon0"] + 360.0,
        )

    if fftpkg == "scipy":
        new_dims = list(fc.dims)
        new_dims[wavenum_ax] = lons.name  # type: ignore
        new_coords = dict(fc.coords)
        new_coords.pop(wave_coord, None)
        new_coords[lons.name] = lons

        filtered = scipy.fft.irfft(
            fc.where(fc[wave_coord].isin(waves), other=0j).values, axis=wavenum_ax
        )
        filtered = xr.DataArray(filtered, dims=new_dims, coords=new_coords)

    elif fftpkg == "xrft":
        filtered = xrft.ifft(
            fc.where(fc[wave_coord].isin(waves), other=0j),
            dim=wave_coord,
            real_dim=wave_coord,
            true_amplitude=False,
            true_phase=False,
            prefix="",
        )
        filtered = filtered.rename({wave_coord: lons.name}).assign_coords(
            {lons.name: lons}
        )

    else:
        msg = "Invalid fftpkg: must be 'scipy' or 'xrft'"
        raise ValueError(msg)

    return filtered


def zonal_wave_contributions(
    fc: xr.DataArray,
    waves: Iterable[int],
    fftpkg: str = "scipy",
    wave_coord: str = "",
    lons: None | xr.DataArray = None,
) -> xr.DataArray:
    r"""Computes the individual contributions of each zonal wavenumber
    to the input field. This is done by taking an inverse FFT of the
    Fourier coefficients, with all but one wavenumber set to zero.

    Parameters
    ----------
    fc : `xarray.DataArray`
        Fourier coefficients as a function of zonal wavenumber, as
        returned by `zonal_wave_coeffs`.
    waves : iterable of int, optional
        The zonal wavenumbers to maintain in the output.
    fftpkg : string, optional
        String that specifies how to perform the FFT on the data. Options
        are scipy or xrft. Specifying scipy uses some operations that are
        memory-eager and leverages scipy.fft.rfft. Specifying xrft should
        leverage the benefits of xarray/dask for large datasets by using
        xrft.fft. Defaults to scipy.
    wave_coord : str, optional
        The coordinate name of the wavenumber dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the wavenumbers
    lons: `xarray.DataArray`, optional
        The longitude coordinate of the input field. If not given, the
        function will attempt to reconstruct the coordinate from the input
        Fourier coefficients, by looking for 'nlons' and 'lon0' attributes.

    Returns
    -------
    `xarray.DataArray`
        The individual contributions of each wavenumber to the original
        input field.

    See Also
    --------
    zonal_wave_coeffs
    filter_by_zonal_wave_truncation

    """

    if wave_coord == "":
        coords = infer_xr_coord_names(fc, required=["zonal_wavenum"])
        wave_coord = coords["zonal_wavenum"]

    contributions = []
    for wave in waves:
        contributions.append(
            filter_by_zonal_wave_truncation(
                fc, [wave], fftpkg=fftpkg, wave_coord=wave_coord, lons=lons
            )
        )
    contributions = xr.concat(contributions, dim=wave_coord).assign_coords(
        {wave_coord: waves}
    )

    return contributions


def zonal_wave_covariance(
    fc1: xr.DataArray,
    fc2: xr.DataArray,
    wave_coord: str = "",
    nlons: int | None = None,
) -> xr.DataArray:
    r"""Calculates the covariance of two fields partititioned into
    zonal wavenumbers.

    Parameters
    ----------
    fc1 : `xarray.DataArray`
        The zonal Fourier coefficients of the first field, as returned by
        `zonal_wave_coeffs`.
    fc2 : `xarray.DataArray`
        The zonal Fourier coefficients of the second field, as returned by
        `zonal_wave_coeffs`.
    wave_coord : str optional
        The coordinate name of the wavenumber dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the wavenumbers.
    nlons: int, optional
        The number of longitudes in the original data. If not provided, the
        function will try to find this in the attrs of fc1 or fc2.

    Returns
    -------
    `xarray.DataArray`

    See Also
    --------
    zonal_wave_coeffs

    """

    # Ensure that input DataArrays are fully consistent
    xr.align(fc1, fc2, join="exact", copy=False)

    # If both DataArrays are fully consistent, then this block will cover both
    if wave_coord == "":
        coords = infer_xr_coord_names(fc1, required=["zonal_wavenum"])
        wave_coord = coords["zonal_wavenum"]

    if (nlons is None) and ("nlons" not in fc1.attrs) and ("nlons" not in fc2.attrs):
        raise ValueError(
            "nlons must either be provided as a kwarg or be in the attrs of fc1 or fc2"
        )
    elif nlons is None:
        nlons = fc1.attrs.get("nlons", fc2.attrs.get("nlons"))

    mult_mask = np.isfinite(fc1.where(fc1.zonal_wavenum != 0)) + 1
    cov = mult_mask * np.real(fc1 * fc2.conj()) / (nlons * nlons)  # type: ignore

    return cov



from __future__ import annotations

import numpy as np
import xarray as xr

from .checks import has_global_regular_lons, infer_xr_coord_names


def zonal_mean(
    dat: xr.DataArray | xr.Dataset,
    lon_coord: str = "",
    strict: bool = False,
) -> xr.DataArray | xr.Dataset:
    r"""Compute the zonal mean.

    This is primarily a convenience function that will make other
    code more explicit/readable. This function is imported at the
    top level of the package by default.

    Parameters
    ----------
    dat : `xarray.DataArray` or `xarray.Dataset`
        data containing a dimension named longitude that spans all 360 degrees

    lon_coord : str, optional
        The coordinate name of the longitude dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the longitude

    strict : bool, optional
        If True, the function will check whether the longitudes span 360
        degrees with regular spacing. If False (the default), this check is
        skipped.

    Returns
    -------
    zonal average: `xarray.DataArray` or `xarray.Dataset`
        The mean across the longitude dimension

    """

    if lon_coord == "":
        coords = infer_xr_coord_names(dat, required=["lon"])
        lon_coord = coords["lon"]

    if strict is True:
        has_global_regular_lons(dat[lon_coord], enforce=True)

    return dat.mean(lon_coord)


def meridional_mean(
    dat: xr.DataArray | xr.Dataset,
    lat1: float,
    lat2: float,
    lat_coord: str = "",
) -> xr.DataArray | xr.Dataset:
    r"""Compute the cos(lat) weighted mean of data between two latitudes.

    This function is imported at the top level of the package by default.

    Parameters
    ----------
    dat : `xarray.DataArray` or `xarray.Dataset`
        data containing a latitude dimension that spans
        lat1 and lat2. The cos(lat) weighting assumes that the
        latitudes are equally spaced. If given a dataset, the
        function assumes all variables are on the same
        latitude grid.

    lat1 : float
        The beginning latitude limit of the band average

    lat2 : float
        The ending latitude limit of the band average

    lat_coord : str, optional
        The coordinate name of the latitude dimension. If given an empty
        string (the default), the function tries to infer which coordinate
        corresponds to the latitude

    Returns
    -------
    meridional average: `xarray.DataArray` or `xarray.Dataset`
        the weighted mean across the latitude dimension limited
        by lat1 and lat2

    """

    if lat_coord == "":
        coords = infer_xr_coord_names(dat, required=["lat"])
        lat_coord = coords["lat"]

    if lat1 >= lat2:
        msg = "lat1 must be less than lat2"
        raise ValueError(msg)

    min_lat = float(dat[lat_coord].min())
    max_lat = float(dat[lat_coord].max())
    if not ((min_lat <= lat1 <= max_lat) and (min_lat <= lat2 <= max_lat)):
        msg = (
            f"data only contains lats in range of {min_lat} to {max_lat} "
            + f"(chose lat1={lat1}, lat2={lat2})"
        )
        raise ValueError(msg)

    lats = dat[lat_coord]
    ixs = {lat_coord: np.logical_and(lats >= lat1, lats <= lat2)}
    wgts = np.cos(np.deg2rad(dat[lat_coord].isel(ixs)))

    return dat.isel(ixs).weighted(wgts).mean(lat_coord)  # type: ignore



import pytest

import xarray as xr
import numpy as np
import scipy

from pyzome.zonal_waves import (
    zonal_wave_coeffs,
    zonal_wave_ampl_phase,
    inflate_zonal_wave_coeffs,
    filter_by_zonal_wave_truncation,
    zonal_wave_contributions,
    zonal_wave_covariance,
)
from pyzome.mock_data import create_dummy_geo_field, lon_coord, lat_coord


def test_zonal_wave_coeffs_scipy():
    """Test that zonal wave coeffs works correctly with scipy FFT"""
    test_lons = lon_coord(2.0)
    da = create_dummy_geo_field(test_lons, lat_coord(3.0))
    da_zonal_wave_coeffs = zonal_wave_coeffs(da, fftpkg="scipy")
    scipy_fft = scipy.fft.rfft(da.values, axis=1)

    # Basic tests that the function works as expected
    assert da_zonal_wave_coeffs.shape == (61, 91)
    assert da_zonal_wave_coeffs.dims == ("lat", "zonal_wavenum")
    assert da_zonal_wave_coeffs.lon0 == test_lons[0]
    assert da_zonal_wave_coeffs.nlons == test_lons.size
    np.testing.assert_allclose(da_zonal_wave_coeffs.values, scipy_fft)

    # Selecting specific wave numbers
    da_zonal_wave_coeffs = zonal_wave_coeffs(da, waves=[1, 2, 3, 4], fftpkg="scipy")
    assert da_zonal_wave_coeffs.shape == (61, 4)
    np.testing.assert_allclose(da_zonal_wave_coeffs.values, scipy_fft[:, 1:5])


def test_zonal_wave_coeffs_xrft():
    """Test that zonal wave coeffs works correctly with xrft FFT"""
    test_lons = lon_coord(2.0)
    da = create_dummy_geo_field(test_lons, lat_coord(3.0))
    da_zonal_wave_coeffs = zonal_wave_coeffs(da, fftpkg="xrft")
    numpy_fft = np.fft.rfft(da.values, axis=1)

    # Basic tests that the function works as expected
    assert da_zonal_wave_coeffs.shape == (61, 91)
    assert da_zonal_wave_coeffs.dims == ("lat", "zonal_wavenum")
    assert da_zonal_wave_coeffs.lon0 == test_lons[0]
    assert da_zonal_wave_coeffs.nlons == test_lons.size
    np.testing.assert_allclose(da_zonal_wave_coeffs.values, numpy_fft)

    # Selecting specific wave numbers
    da_zonal_wave_coeffs = zonal_wave_coeffs(da, waves=[1, 2, 3, 4], fftpkg="xrft")
    assert da_zonal_wave_coeffs.shape == (61, 4)
    np.testing.assert_allclose(da_zonal_wave_coeffs.values, numpy_fft[:, 1:5])


def test_zonal_wave_coeffs_scipy_xrft_equal():
    """Test that zonal wave coeffs give same results with scipy and xrft"""
    test_lons = lon_coord(2.0)
    da = create_dummy_geo_field(test_lons, lat_coord(3.0))
    da_fc_xrft = zonal_wave_coeffs(da, fftpkg="xrft").astype("complex64")
    da_fc_scipy = zonal_wave_coeffs(da, fftpkg="scipy")

    assert da_fc_xrft.shape == da_fc_scipy.shape
    assert da_fc_xrft.dims == da_fc_scipy.dims
    assert da_fc_xrft.lon0 == da_fc_scipy.lon0
    assert da_fc_xrft.nlons == da_fc_scipy.nlons
    np.testing.assert_allclose(da_fc_xrft.values, da_fc_scipy.values, atol=1.0e-5)


def test_zonal_wave_coeffs_unknown_fftpkg():
    """Test that zonal wave coeffs fails with unknown fftpkg"""
    da = create_dummy_geo_field(lon_coord(10), lat_coord(10))
    with pytest.raises(ValueError) as e:
        da_zonal_wave_coeffs = zonal_wave_coeffs(da, fftpkg="unknown")
    assert "fftpkg must be 'scipy' or 'xrft'" in str(e.value)


def test_inflate_zonal_wave_coeffs():
    """Test that inflate_zonal_wave_coeffs works correctly"""
    da = create_dummy_geo_field(lon_coord(2.0), lat_coord(3.0))

    da_fc = zonal_wave_coeffs(da, fftpkg="scipy", waves=[1, 2, 3, 4])
    da_fc_inflated = inflate_zonal_wave_coeffs(da_fc)
    assert da_fc_inflated.shape == (61, 91)
    assert np.allclose(da_fc_inflated.sel(zonal_wavenum=[1, 2, 3, 4]), da_fc)

    # Test the case when non-consecutive wavenums are initially kept
    da_fc = zonal_wave_coeffs(da, fftpkg="scipy", waves=[1, 5, 10, 20])
    da_fc_inflated = inflate_zonal_wave_coeffs(da_fc)
    assert np.allclose(da_fc_inflated.sel(zonal_wavenum=[1, 5, 10, 20]), da_fc)


def test_inflate_zonal_wave_coeffs_missing_nlons():
    """Test that inflate_zonal_wave_coeffs fails without nlons attribute"""
    da = create_dummy_geo_field(lon_coord(2.0), lat_coord(3.0))

    da_fc = zonal_wave_coeffs(da, fftpkg="scipy", waves=[1, 2, 3, 4])
    _ = da_fc.attrs.pop("nlons")

    with pytest.raises(KeyError) as e:
        da_fc_inflated = inflate_zonal_wave_coeffs(da_fc)
    assert "input DataArray must have an 'nlons' attribute" in str(e.value)


def test_inflate_zonal_wave_coeffs_return_input():
    """Test that inflate_zonal_wave_coeffs returns input if it matches desired wavenums"""
    da = create_dummy_geo_field(lon_coord(2.0), lat_coord(3.0))

    da_fc = zonal_wave_coeffs(da, fftpkg="scipy")

    da_fc_inflated = inflate_zonal_wave_coeffs(da_fc)
    assert da_fc_inflated is da_fc


def test_inflate_zonal_wave_coeffs_unexpected_wavenums():
    """Test that inflate_zonal_wave_coeffs fails with unexpected wavenums"""
    da = create_dummy_geo_field(lon_coord(2.0), lat_coord(3.0))

    da_fc = zonal_wave_coeffs(da, fftpkg="scipy", waves=[1, 2, 3, 4])
    da_fc = da_fc.assign_coords({"zonal_wavenum": [1, 2, 103, 104]})

    with pytest.raises(ValueError) as e:
        da_fc_inflated = inflate_zonal_wave_coeffs(da_fc)
    assert (
        "input DataArray wavenumbers are not a subset of the expected wavenumbers based on the 'nlons' attribute"
        in str(e.value)
    )


def test_zonal_wave_ampl_phase():
    """Test that zonal wave ampl phase works correctly with scipy FFT"""
    test_lons = lon_coord(1.0)
    test_lats = lat_coord(1.0)

    wave_num = 3
    wave_ampl = 4
    wave_phase = -np.pi / 3

    # Test with a cosine wave
    wave_field = np.broadcast_to(
        4 * np.cos(wave_num * np.deg2rad(test_lons) + wave_phase),
        (test_lats.size, test_lons.size),
    )
    wave_field = xr.DataArray(
        wave_field, coords=[test_lats, test_lons], dims=["lat", "lon"]
    )
    wave_field_fc = zonal_wave_coeffs(wave_field, fftpkg="scipy")

    ampl, phase = zonal_wave_ampl_phase(wave_field_fc)

    assert np.allclose(ampl.sel(zonal_wavenum=wave_num, lat=0), wave_ampl)
    assert np.allclose(phase.sel(zonal_wavenum=wave_num, lat=0), wave_phase)


def test_zonal_wave_ampl_phase_missing_nlons():
    da = create_dummy_geo_field(lon_coord(2.0), lat_coord(3.0))
    da_fc = zonal_wave_coeffs(da, fftpkg="scipy", waves=[1, 2, 3])

    with pytest.raises(KeyError) as e:
        _ = da_fc.attrs.pop("nlons")
        ampl, phase = zonal_wave_ampl_phase(da_fc)
    assert "input DataArray must have an 'nlons' attribute" in str(e.value)


def test_filter_by_zonal_wave_truncation_scipy():
    """Test that zonal wave truncation filtering works correctly with scipy IFFT"""
    test_lons = lon_coord(1.0)
    test_lats = lat_coord(1.0)

    wave_fields = [
        np.broadcast_to(
            k * np.cos(k * np.deg2rad(test_lons)), (test_lats.size, test_lons.size)
        )
        for k in range(0, 5)
    ]
    wave_field_13 = wave_fields[1] + wave_fields[3]
    wave_field_24 = wave_fields[2] + wave_fields[4]
    wave_field_134 = wave_fields[1] + wave_fields[3] + wave_fields[4]

    full_wave_field = xr.DataArray(
        wave_fields[1] + wave_fields[2] + wave_fields[3] + wave_fields[4],
        coords=[test_lats, test_lons],
        dims=["lat", "lon"],
    )
    wave_field_fc = zonal_wave_coeffs(full_wave_field, fftpkg="scipy")

    wave_field_filtered = filter_by_zonal_wave_truncation(
        wave_field_fc, [1, 3], fftpkg="scipy"
    )
    np.testing.assert_allclose(wave_field_filtered.values, wave_field_13, atol=1.0e-12)

    wave_field_filtered = filter_by_zonal_wave_truncation(
        wave_field_fc, [2, 4], fftpkg="scipy"
    )
    np.testing.assert_allclose(wave_field_filtered.values, wave_field_24, atol=1.0e-12)

    wave_field_filtered = filter_by_zonal_wave_truncation(
        wave_field_fc, [1, 3, 4], fftpkg="scipy"
    )
    np.testing.assert_allclose(wave_field_filtered.values, wave_field_134, atol=1.0e-12)


def test_filter_by_zonal_wave_truncation_xrft():
    """Test that zonal wave truncation filtering works correctly with xrft IFFT"""
    test_lons = lon_coord(1.0)
    test_lats = lat_coord(1.0)

    wave_fields = [
        np.broadcast_to(
            k * np.cos(k * np.deg2rad(test_lons)), (test_lats.size, test_lons.size)
        )
        for k in range(0, 5)
    ]
    wave_field_13 = wave_fields[1] + wave_fields[3]
    wave_field_24 = wave_fields[2] + wave_fields[4]
    wave_field_134 = wave_fields[1] + wave_fields[3] + wave_fields[4]

    full_wave_field = xr.DataArray(
        wave_fields[1] + wave_fields[2] + wave_fields[3] + wave_fields[4],
        coords=[test_lats, test_lons],
        dims=["lat", "lon"],
    )
    wave_field_fc = zonal_wave_coeffs(full_wave_field, fftpkg="xrft")

    wave_field_filtered = filter_by_zonal_wave_truncation(
        wave_field_fc, [1, 3], fftpkg="xrft"
    )
    np.testing.assert_allclose(wave_field_filtered.values, wave_field_13, atol=1.0e-12)

    wave_field_filtered = filter_by_zonal_wave_truncation(
        wave_field_fc, [2, 4], fftpkg="xrft"
    )
    np.testing.assert_allclose(wave_field_filtered.values, wave_field_24, atol=1.0e-12)

    wave_field_filtered = filter_by_zonal_wave_truncation(
        wave_field_fc, [1, 3, 4], fftpkg="xrft"
    )
    np.testing.assert_allclose(wave_field_filtered.values, wave_field_134, atol=1.0e-12)


def test_filter_by_zonal_wave_truncation_no_attrs():
    da = create_dummy_geo_field(lon_coord(2.0), lat_coord(3.0))
    da_fc = zonal_wave_coeffs(da, fftpkg="scipy")

    with pytest.raises(KeyError) as e:
        _ = da_fc.attrs.pop("lon0")
        da_filt = filter_by_zonal_wave_truncation(da_fc, [1, 2, 3])
    assert "input DataArray must have a 'lon0' attribute" in str(e.value)

    with pytest.raises(KeyError) as e:
        _ = da_fc.attrs.pop("nlons")
        da_filt = filter_by_zonal_wave_truncation(da_fc, [1, 2, 3])
    assert "input DataArray must have an 'nlons' attribute" in str(e.value)


def test_filter_by_zonal_wave_truncation_infer_lons():
    """Test that filter_by_zonal_wave_truncation can infer longitudes correctly"""
    test_lons = lon_coord(2.0)
    da = create_dummy_geo_field(test_lons, lat_coord(2.0))
    da_fc = zonal_wave_coeffs(da, fftpkg="scipy")

    with pytest.warns(UserWarning) as w:
        da_filtered = filter_by_zonal_wave_truncation(
            da_fc, [1, 6], fftpkg="scipy", lons=None
        )
    assert "attempting to infer the input longitude coordinate" in str(w[0].message)
    np.testing.assert_allclose(da_filtered.lon.values, test_lons)


def test_filter_by_zonal_wave_truncation_fftpkg_fails():
    """Test that filter_by_zonal_wave_truncation fails with invalid fftpkg"""
    test_lons = lon_coord(2.0)
    da = create_dummy_geo_field(test_lons, lat_coord(2.0))
    da_fc = zonal_wave_coeffs(da, fftpkg="scipy")

    with pytest.raises(ValueError) as e:
        da_filt = filter_by_zonal_wave_truncation(da_fc, [1, 2, 3], fftpkg="foo")
    assert "Invalid fftpkg: must be" in str(e.value)


def test_zonal_wave_contributions():
    """Test that zonal wave truncation filtering works correctly with scipy IFFT"""
    test_lons = lon_coord(1.0)
    test_lats = lat_coord(1.0)

    wave_fields = [
        np.broadcast_to(
            k * np.cos(k * np.deg2rad(test_lons)), (test_lats.size, test_lons.size)
        )
        for k in range(0, 5)
    ]

    full_wave_field = xr.DataArray(
        wave_fields[1] + wave_fields[2] + wave_fields[3] + wave_fields[4],
        coords=[test_lats, test_lons],
        dims=["lat", "lon"],
    )
    wave_field_fc = zonal_wave_coeffs(full_wave_field, fftpkg="scipy")
    contributions = zonal_wave_contributions(
        wave_field_fc,
        [1, 2, 3, 4],
        lons=test_lons,
    )

    for k in range(1, 5):
        np.testing.assert_allclose(
            contributions.sel(zonal_wavenum=k).values, wave_fields[k], atol=1.0e-12
        )


def test_zonal_wave_covariance():
    """Test that zonal wave covariance works correctly"""
    test_lons = lon_coord(1.0)
    test_lats = lat_coord(2.0)

    wave1 = xr.DataArray(
        np.broadcast_to(
            np.cos(np.deg2rad(test_lons)), (test_lats.size, test_lons.size)
        ),
        coords=[test_lats, test_lons],
        dims=["lat", "lon"],
    )
    wave1_phased = xr.DataArray(
        np.broadcast_to(
            np.cos(np.deg2rad(test_lons) + np.pi / 3), (test_lats.size, test_lons.size)
        ),
        coords=[test_lats, test_lons],
        dims=["lat", "lon"],
    )

    fc_k1 = zonal_wave_coeffs(wave1, fftpkg="scipy")
    fc_k1_phased = zonal_wave_coeffs(wave1_phased, fftpkg="scipy")

    zon_cov = zonal_wave_covariance(fc_k1, fc_k1_phased)

    assert zon_cov.dims == ("lat", "zonal_wavenum")
    np.testing.assert_allclose(zon_cov.sel(zonal_wavenum=1).values, 0.25)
    np.testing.assert_allclose(
        zon_cov.isel(zonal_wavenum=slice(2, None)).values, 0.0, rtol=0, atol=1.0e-30
    )


def test_zonal_wave_covariance_no_nlons():
    """Test that zonal_wave_covariance fails without any nlons given"""

    da1 = create_dummy_geo_field(lon_coord(2.5), lat_coord(2.5))
    da2 = create_dummy_geo_field(lon_coord(2.5), lat_coord(2.5))
    da_fc1 = zonal_wave_coeffs(da1, fftpkg="scipy")
    da_fc2 = zonal_wave_coeffs(da2, fftpkg="scipy")

    with pytest.raises(ValueError) as e:
        _ = da_fc1.attrs.pop("nlons")
        _ = da_fc2.attrs.pop("nlons")

        zon_cov = zonal_wave_covariance(da_fc1, da_fc2)
    assert (
        "nlons must either be provided as a kwarg or be in the attrs of fc1 or fc2"
        in str(e.value)
    )



import pytest

from pyzome.mock_data import create_dummy_geo_dataset, lon_coord, lat_coord, plev_coord
from pyzome.recipes.zmd import create_zonal_mean_dataset


def test_create_zonal_mean_dataset_no_waves():
    """Test that create_zonal_mean_dataset works when no wave-output is requested"""
    ds = create_dummy_geo_dataset(
        field_names=["u", "v", "w", "T", "Z"],
        lons=lon_coord(2.5),
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
    )

    zmds = create_zonal_mean_dataset(ds, include_waves=False)
    assert "zonal_wavenum" not in zmds.coords
    for var in ["u", "v", "w", "T", "Z", "uv", "vT", "uw", "wT"]:
        assert var in zmds.data_vars
        assert "units" in zmds[var].attrs
        assert "long_name" in zmds[var].attrs

    ds = ds.drop_vars(["u", "Z"])
    zmds = create_zonal_mean_dataset(ds, include_waves=False)
    for var in ["v", "w", "T", "vT", "wT"]:
        assert var in zmds.data_vars
    for var in ["u", "uv", "uw"]:
        assert var not in zmds.data_vars


def test_create_zonal_mean_dataset_with_waves():
    """Test that create_zonal_mean_dataset works when wave-output is requested"""
    ds = create_dummy_geo_dataset(
        field_names=["u", "v", "w", "T", "Z"],
        lons=lon_coord(2.5),
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
    )

    zmds = create_zonal_mean_dataset(ds, include_waves=True, waves=[1, 2, 3])
    assert "zonal_wavenum" in zmds.coords

    expected_vars = [
        "u",
        "v",
        "w",
        "T",
        "Z",
        "uv",
        "vT",
        "uw",
        "wT",
        "uv_k",
        "vT_k",
        "uw_k",
        "wT_k",
        "Z_k_real",
        "Z_k_imag",
        "T_k_real",
        "T_k_imag",
    ]
    for var in expected_vars:
        assert var in zmds.data_vars
        assert "units" in zmds[var].attrs
        assert "long_name" in zmds[var].attrs

    ds = ds.drop_vars(["u", "Z"])
    zmds = create_zonal_mean_dataset(ds, include_waves=True, waves=[1, 2, 3])

    expected_vars = ["v", "w", "T", "vT", "wT", "vT_k", "wT_k", "T_k_real", "T_k_imag"]
    for var in expected_vars:
        assert var in zmds.data_vars

    expected_missing = ["u", "uv", "uw", "uv_k", "uw_k", "Z_k_real", "Z_k_imag"]
    for var in expected_missing:
        assert var not in zmds.data_vars


def test_create_zonal_mean_dataset_fails_no_valid_fields():
    """Test that create_zonal_mean_dataset fails when no valid fields are provided"""
    ds = create_dummy_geo_dataset(
        field_names=["foo", "bar", "baz"],
        lons=lon_coord(2.5),
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
    )

    with pytest.raises(ValueError) as e:
        zmds = create_zonal_mean_dataset(ds)
    assert "No valid fields found in provided dataset" in str(e.value)



import pytest

import numpy as np

from pyzome.checks import (
    has_global_regular_lons,
    infer_xr_coord_names,
    check_var_SI_units,
    check_for_logp_coord,
    var_units,
)
from pyzome.exceptions import AttrError, CoordinateError, LongitudeError, UnitError
from pyzome.mock_data import create_dummy_geo_field, lon_coord, lat_coord, plev_coord


def test_global_regular_lons():
    """Test that longitudes span globe and are evenly spaced"""

    # Full 360 span from 0 to 360
    lons_full_360 = np.arange(0, 360, 2.0)
    assert has_global_regular_lons(lons_full_360, enforce=True)

    # Full 360 span from -180 to 180
    lons_full_180 = np.arange(-180, 180, 3.0)
    assert has_global_regular_lons(lons_full_180, enforce=True)


def test_partial_lons_fail():
    """Test that partial longitude coverage fails"""

    # Partial span from 45 to 225
    lons_part_equal = np.arange(45, 225, 5.0)
    assert has_global_regular_lons(lons_part_equal, enforce=False) is False
    with pytest.raises(LongitudeError) as e:
        has_global_regular_lons(lons_part_equal, enforce=True)
    assert "longitudes do not span all 360 degrees" in str(e.value)


def test_irregular_lons_fail():
    """Test that irregularly spaced longitudes fail"""

    # Full span with unequal spacing
    lons_full_unequal = np.array([0, 20, 40, 60, 90, 120, 150, 180, 225, 270, 315, 359])
    assert has_global_regular_lons(lons_full_unequal, enforce=False) is False
    with pytest.raises(LongitudeError) as e:
        has_global_regular_lons(lons_full_unequal, enforce=True)
    assert "longitudes are not equally spaced" in str(e.value)


def test_infer_xr_coord_names():
    """Test that standard coordinates can be inferred/detected"""

    da = create_dummy_geo_field(lon_coord(10), lat_coord(10), plev_coord(3))
    mapper = {"lev": "lev", "lat": "lat", "lon": "lon"}
    assert infer_xr_coord_names(da, required=["lev", "lat", "lon"]) == mapper

    da = create_dummy_geo_field(
        lon_coord(10, name="longitude"),
        lat_coord(10, name="latitude"),
        plev_coord(3, name="level"),
    )
    print(da)
    mapper = {"lev": "level", "lat": "latitude", "lon": "longitude"}
    assert infer_xr_coord_names(da, required=["lev", "lat", "lon"]) == mapper


def test_infer_multiple_coords_fail():
    """Test detection of multiple ambiguous coordinate names"""

    da = create_dummy_geo_field(lon_coord(10), lon_coord(10, name="LONGI"))
    with pytest.raises(CoordinateError) as e:
        infer_xr_coord_names(da, required=["lon"])
    assert "Found multiple coordinates in dat matching the" in str(e.value)


def test_infer_coords_missing_fail():
    """Test that a required coordinate missing is detected"""

    da = create_dummy_geo_field(
        lon_coord(10, name="notgitude"), lat_coord(10, name="latitude")
    )
    with pytest.raises(CoordinateError) as e:
        infer_xr_coord_names(da, required=["lon"])
    assert "Unable to match any of the coordinates in dat for" in str(e.value)


def test_check_var_SI_units():
    """Test that all relevant SI units can be detected"""

    da = create_dummy_geo_field(lon_coord(10), lat_coord(10))
    for var in var_units:
        for unit in var_units[var]:
            da.attrs["units"] = unit
        assert check_var_SI_units(da, var, enforce=True)


def test_check_var_SI_units_missing():
    """Test that missing units attribute throws an error"""

    da = create_dummy_geo_field(lon_coord(10), lat_coord(10))
    with pytest.raises(AttrError) as e:
        check_var_SI_units(da, "wind", enforce=False)
    assert "units is not an attribute of the given DataArray" in str(e.value)


def test_check_SI_units_wrong():
    """Test that wrong/non-SI units are detected"""

    da = create_dummy_geo_field(lon_coord(10), lat_coord(10), attrs={"units": "N/A"})

    assert check_var_SI_units(da, "temperature", enforce=False) is False
    with pytest.raises(UnitError) as e:
        check_var_SI_units(da, "pressure", enforce=True)
    assert "do not match SI units for the" in str(e.value)


def test_check_for_logp_coord():
    """Test that check_for_logp_coord properly detects a log-p altitude coordinate"""
    da = create_dummy_geo_field(
        lon_coord(10), 
        lat_coord(10), 
        plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
    )

    assert check_for_logp_coord(da) is False 
    with pytest.raises(CoordinateError) as e:
        check_for_logp_coord(da, enforce=True)
    assert "z is not a coordinate in the data" in str(e.value) 

    da = da.assign_coords({"z": np.log(da.lev)})
    assert check_for_logp_coord(da) is False
    with pytest.raises(AttrError) as e:
        check_for_logp_coord(da, enforce=True)
    assert "z is missing at least one of the required attributes" in str(e.value)

    da.z.attrs["long_name"] = "not the right name"
    da.z.attrs["units"] = "not the right units"
    da.z.attrs["note"] = "not the right note"
    assert check_for_logp_coord(da) is False
    with pytest.raises(AttrError) as e:
        check_for_logp_coord(da, enforce=True)
    assert "z must have a long_name = 'log-pressure altitude'" in str(e.value)

    da.z.attrs["long_name"] = "log-pressure altitude"
    assert check_for_logp_coord(da) is False
    with pytest.raises(AttrError) as e:
        check_for_logp_coord(da, enforce=True)
    assert "z must have a note = 'added by pyzome'" in str(e.value)

    da.z.attrs["note"] = "added by pyzome"
    assert check_for_logp_coord(da) is False
    with pytest.raises(UnitError) as e:
        check_for_logp_coord(da, enforce=True)
    assert "do not match SI units for the" in str(e.value)

    da.z.attrs["units"] = "m"
    assert check_for_logp_coord(da, enforce=True) is True


    


class Error(Exception):
    pass


class LongitudeError(Error):
    """Exception raised when the longitude coordinate of a Dataset/DataArray
    doesn't span the globe or when it is irregularly spaced"""

    def __init__(self, message):
        self.message = message


class CoordinateError(Error):
    """Exception raised when the input Dataset/DataArray doesn't have the
    appropriate coordinate(s)."""

    def __init__(self, message):
        self.message = message


class AttrError(Error):
    """Exception raised when the input Dataset/DataArray doesn't have the
    appropriate attribute(s)."""

    def __init__(self, message):
        self.message = message


class UnitError(Error):
    """Exception raised when the input DataArray doesn't have the
    appropriate units string for the given variable."""

    def __init__(self, message):
        self.message = message



from __future__ import annotations

import numpy as np
import xarray as xr


def _make_regular_coord(
    left: float | int,
    right: float | int,
    step: float | int,
    name: str,
    inclusive: bool = False,
) -> xr.DataArray:
    """
    Generate an xr.DataArray representing a regularly spaced coordinate array

    Parameters
    ----------
    left : float | int
        The left limit of the coordinate
    right : float | int
        The right limit of the coordinate
    step : float | int
        The step size between each element of the coordinate
    name : str
        The name of the coordinate
    inclusive : bool, optional
        Whether to include the right limit in the coordinate

    Returns
    -------
    `xarray.DataArray`

    """
    if step == 0.0:
        msg = "step cannot be zero"
        raise ValueError(msg)

    if left == right:
        msg = "left- and right-side values cannot be equal"
        raise ValueError(msg)

    if ((left < right) and (step < 0)) or ((left > right) and (step > 0)):
        msg = f"Invalid step for the given limits: left={left}, right={right}, step={step}"
        raise ValueError(msg)

    coord = np.arange(left, right + (inclusive * step), step).astype("float64")
    return xr.DataArray(coord, dims=[name], name=name).assign_coords({name: coord})


def lat_coord(
    resolution: float | int,
    name: str = "lat",
    left_lim: float | int = -90.0,
    right_lim: float | int = 90.0,
    inclusive: bool = True,
) -> xr.DataArray:
    """
    Generate a regularly spaced latitude-like coordinate

    Parameters
    ----------
    resolution : float | int
        The spacing between latitudes in degrees
    name : str, optional
        The name of the coordinate, defaults to "lat"
    left_lim : float | int, optional
        The left latitude limit, which defaults to -90
    right_lim : float | int, optional
        The right latitude limit, which defaults to 90
    inclusive : bool, optional
        Whether to include the right latitude limit, which defaults to True

    Returns
    -------
    latitudes: `xarray.DataArray` 
        Regularly spaced latitudes with the given resolution 
        (and name/limits, if given)

    """
    if (abs(left_lim) > 90) or (abs(right_lim) > 90):
        msg = f"Latitudes cannot be below -90 or above 90: left_lim={left_lim}, upp_lim={right_lim}"
        raise ValueError(msg)

    lats = _make_regular_coord(
        left_lim, right_lim, resolution, name, inclusive=inclusive
    )
    lats.attrs["standard_name"] = "latitude"
    lats.attrs["units"] = "degrees_north"

    return lats


def lon_coord(
    resolution: float | int,
    name: str = "lon",
    left_lim: float | int = 0.0,
    right_lim: float | int = 360.0,
    inclusive: bool = False,
) -> xr.DataArray:
    """
    Generate a regularly spaced longitude-like coordinate

    Parameters
    ----------
    resolution : float | int
        The spacing between longitudes in degrees
    name : str, optional
        The name of the coordinate, defaults to "lon"
    left_lim : float | int, optional
        The left longitude limit, which defaults to 0
    right_lim : float | int, optional
        The right longitude limit, which defaults to 360
    inclusive : bool, optional
        Whether to include the right longitude limit, which defaults to False

    Returns
    -------
    longitudes: `xarray.DataArray` 
        Regularly spaced longitudes with the given resolution
        (and name/limits, if given)

    """

    if ((left_lim < 0) and (right_lim > 180)) or ((right_lim < 0) and (left_lim > 180)):
        msg = "Proper longitude range should be within [-180,180] or [0, 360]"
        raise ValueError(msg)
    elif (left_lim > 360) or (right_lim > 360):
        msg = "upper longitude limit cannot exceed 360"
        raise ValueError(msg)
    elif (left_lim < -180) or (right_lim < -180):
        msg = "lower longitude limit cannot be below -180"
        raise ValueError(msg)

    lons = _make_regular_coord(
        left_lim, right_lim, resolution, name, inclusive=inclusive
    )
    lons.attrs["standard_name"] = "longitude"
    lons.attrs["units"] = "degrees_east"

    return lons


def plev_coord(
    levels_per_decade: int,
    left_lim_exponent: int = 3,
    right_lim_exponent: int = 0,
    name: str = "lev",
    units: str = "hPa",
) -> xr.DataArray:
    """
    Generate a pressure-like vertical coordinate with logarithmic spacing

    Parameters
    ----------
    levels_per_decade : int
        The number of levels per decade (i.e., between 10^3 and 10^2, 10^2 
        and 10^1, etc.)
    left_lim_exponent : int, optional
        The base-10 exponent for the left-most pressure level, which defaults 
        to 3 (for 1000 hPa)
    right_lim_exponent : int, optional
        The base-10 exponent for the right-most pressure level, which defaults 
        to 0 (for 1 hPa)
    name : str, optional
        The name of the coordinate, which defaults to "lev"
    units : str, optional
        The units of the coordinate, which defaults to "hPa"

    Returns
    -------
    pressures: `xarray.DataArray`
        Pressure-like coordinate with logarithmic spacing having the given
        number of levels per decade (and name/limits/units, if given)

    """

    if levels_per_decade <= 0:
        raise ValueError(
            f"'levels_per_decade' must be greater than 0, got {levels_per_decade}"
        )
    if left_lim_exponent == right_lim_exponent:
        raise ValueError(
            "'left_lim_exponent' and 'right_lim_exponent' must not be the same"
        )

    num_elements = (
        levels_per_decade * np.abs(left_lim_exponent - right_lim_exponent) + 1
    )
    pres = np.logspace(left_lim_exponent, right_lim_exponent, num_elements)
    pres = xr.DataArray(pres, dims=[name], name=name).assign_coords({name: pres})
    pres.name = name
    pres.attrs["standard_name"] = "pressure"
    pres.attrs["units"] = units

    return pres


def time_coord(
    start: str = "2000-01-01",
    end: str = "2001-01-01",
    freq: str = "M",
) -> xr.DataArray:
    """
    Generate a time-like coordinate

    Parameters
    ----------
    start : str, optional
        The start date, which defaults to "2000-01-01"
    end : str, optional
        The end date, which defaults to "2001-01-01"
    freq : str, optional
        The frequency of the time coordinate, which defaults to "M" (monthly).
        See https://numpy.org/doc/stable/reference/arrays.datetime.html#basic-datetimes
        for more information.

    Returns
    -------
    times: `xarray.DataArray`
        Time-like coordinate with the given start/end dates and frequency

    """

    if start >= end:
        raise ValueError("start time must be before end time")

    times = np.arange(start, end, dtype=f"datetime64[{freq}]").astype("datetime64[ns]")
    times = xr.DataArray(times, dims=["time"], name="time").assign_coords(
        {"time": times}
    )
    times.attrs["standard_name"] = "time"
    times.attrs["units"] = f"days since {start}"

    return times


def create_dummy_geo_field(
    lons: None | xr.DataArray = None,
    lats: None | xr.DataArray = None,
    levs: None | xr.DataArray = None,
    times: None | xr.DataArray = None,
    name: str = "dummy",
    attrs: dict = {},
) -> xr.DataArray:
    """
    Mock a geophysical field with random data

    Parameters
    ----------
    lons : `xarray.DataArray`
        The longitudes of the data. Defaults to None (longitudes excluded)
    lats : `xarray.DataArray`
        The latitudes of the data. Defaults to None (latitudes excluded)
    levs : `xarray.DataArray`, optional
        The vertical levels of the data. Defaults to None (vertical levels excluded)
    times : `xarray.DataArray`, optional
        The times of the data. Defaults to None (times excluded)
    name : str, optional
        The name of the data. Defaults to "dummy"
    attrs : dict, optional
        The attributes of the data. Defaults to an empty dictionary.

    Returns
    -------
    dummy_data: `xarray.DataArray` 
        Random data with corresponding geophysical coordinates

    """

    dat_shape = []
    coords = []
    dims = []

    for coord in (lons, lats, levs, times):
        if coord is not None:
            dat_shape.insert(0, coord.size)
            coords.insert(0, coord)
            dims.insert(0, coord.name)

    if len(dat_shape) == 0:
        msg = "At least one of 'lons', 'lats', 'levs', or 'times' must be provided"
        raise ValueError(msg)

    dummy = np.random.normal(size=dat_shape).astype("float32")
    da = xr.DataArray(dummy, coords=coords, dims=dims, name=name)
    da.attrs = attrs
    return da


def create_dummy_geo_dataset(
    field_names: list[str],
    lons: None | xr.DataArray = None,
    lats: None | xr.DataArray = None,
    levs: None | xr.DataArray = None,
    times: None | xr.DataArray = None,
    field_attrs: None | dict = None,
) -> xr.Dataset:
    """
    Mock a dataset of geophysical fields with random data

    Parameters
    ----------
    field_names : list[str]
        The names of the fields
    lons : `xarray.DataArray`
        The longitudes of the data
    lats : `xarray.DataArray`
        The latitudes of the data
    levs : `xarray.DataArray`, optional
        The vertical levels of the data. Defaults to None (vertical levels excluded)
    times : `xarray.DataArray`, optional
        The times of the data. Defaults to None (times excluded)
    field_attrs : dict, optional
        The attributes of the data. Defaults to None. If provided, this should be
        a nested dictionary with the outer keys corresponding to the field names,
        and inner keys/dicts corresponding to the attributes for each field.
    
    Returns
    -------
    dummy_dataset: `xarray.Dataset` 
        Dataset with fields having random data matching the given geophysical coordinates

    """
    if field_attrs is not None and len(field_attrs) != len(field_names):
        raise ValueError(
            f"Number of field names ({len(field_names)}) must match number of keys in field attributes ({len(field_attrs)})"
        )

    ds = xr.Dataset()
    for name in field_names:
        da = create_dummy_geo_field(lons, lats, levs, times, name)
        ds[name] = da
        if field_attrs is not None:
            ds[name].attrs = field_attrs[name]
    return ds



import pytest
import numpy as np

from pyzome.basic import zonal_mean, meridional_mean
from pyzome.exceptions import AttrError, CoordinateError, LongitudeError, UnitError
from pyzome.mock_data import create_dummy_geo_field, lon_coord, lat_coord, plev_coord


def test_zonal_mean():
    """Test that zonal mean  works correctly"""
    da = create_dummy_geo_field(lon_coord(10), lat_coord(10), plev_coord(3))
    da_zonal_mean = zonal_mean(da, strict=True)

    # Basic tests that the function works as expected
    assert np.allclose(da_zonal_mean.values, np.mean(da.values, axis=2))
    assert da_zonal_mean.shape == (10, 19)
    assert da_zonal_mean.dims == ("lev", "lat")


def test_meridional_mean():
    """Test that meridional mean works correctly"""
    da = create_dummy_geo_field(lon_coord(10), lat_coord(10), plev_coord(3))
    da_meridional_mean = meridional_mean(da, -90, 90)

    # Basic tests that the function works as expected
    global_mean = np.average(
        da.values, axis=1, weights=np.cos(np.deg2rad(da.lat.values))
    )
    assert np.allclose(da_meridional_mean.values, global_mean)
    assert da_meridional_mean.shape == (10, 36)
    assert da_meridional_mean.dims == ("lev", "lon")


def test_meridional_mean_wrong_lat_order_fail():
    """Test that meridional mean fails with lats in wrong order"""
    da = create_dummy_geo_field(lon_coord(10), lat_coord(10), plev_coord(3))
    with pytest.raises(ValueError) as e:
        da_meridional_mean = meridional_mean(da, 90, -90)
    assert "lat1 must be less than lat2" in str(e.value)


def test_meridional_mean_too_few_lats_fail():
    """Test that meridional mean fails with too few lats"""
    da = create_dummy_geo_field(lon_coord(10), lat_coord(10), plev_coord(3))
    with pytest.raises(ValueError) as e:
        da_meridional_mean = meridional_mean(da.sel(lat=slice(30, 60)), 0, 90)
    assert "data only contains lats in range of" in str(e.value)



from __future__ import annotations
import logging
from typing import Iterable

import xarray as xr

from ..basic import zonal_mean
from ..zonal_waves import zonal_wave_coeffs, zonal_wave_covariance
from ..checks import infer_xr_coord_names

logging.basicConfig(level=logging.NOTSET)

LONG_NAMES = {
    "u": "zonal mean zonal wind",
    "v": "zonal mean meridional wind",
    "w": "zonal mean vertical pressure velocity",
    "T": "zonal mean temperature",
    "Z": "zonal mean geopotential height",
    "uv": "zonal mean eddy momentum flux",
    "vT": "zonal mean eddy heat flux",
    "uw": "zonal mean eddy vertical momentum flux",
    "wT": "zonal mean eddy vertical heat flux",
    "uv_k": "eddy momentum flux due to zonal wavenumber k",
    "vT_k": "eddy heat flux due to zonal wavenumber k",
    "uw_k": "eddy vertical momentum flux due to zonal wavenumber k",
    "wT_k": "eddy vertical heat flux due to zonal wavenumber k",
    "Z_k_real": "real part of zonally fourier transformed geopotential height",
    "Z_k_imag": "imaginary part of zonally fourier transformed geopotential height",
    "T_k_real": "real part of zonally fourier transformed temperature",
    "T_k_imag": "imaginary part of zonally fourier transformed temperature",
}

UNITS = {
    "u": "m s-1",
    "v": "m s-1",
    "w": "Pa s-1",
    "T": "K",
    "Z": "m",
    "uv": "m+2 s-2",
    "vT": "K m s-1",
    "uw": "m Pa s-2",
    "wT": "K Pa s-1",
    "uv_k": "m+2 s-2",
    "vT_k": "K m s-1",
    "uw_k": "m Pa s-2",
    "wT_k": "K Pa s-1",
    "Z_k_real": "m",
    "Z_k_imag": "m",
    "T_k_real": "K",
    "T_k_imag": "K",
}


def create_zonal_mean_dataset(
    ds: xr.Dataset,
    verbose: bool = False,
    include_waves: bool = False,
    waves: None | Iterable[int] = None,
    fftpkg: str = "scipy",
    lon_coord: str = "",
):
    r"""Compiles a "zonal mean dataset".

    Given an xarray dataset containing full fields of basic state
    variables such as velocity components and temperatures, this
    function will compute as many zonal mean diagnostics as possible.

    Parameters
    ----------
    ds : `xarray.Dataset`
        Dataset containing full fields (i.e., containing latitude &
        longitude dimensions) of basic state variables. This function
        currently assumes specific names and units:

        - 'u' = zonal wind component in m/s
        - 'v' = meridional wind component in m/s
        - 'w' = vertical pressure velocity in Pa/s
        - 'T' = temperature in K
        - 'Z' = geopotential height in m

        If your data names, dimensions, and/or units do not conform to these
        restrictions, please change beforehand. Dimensions and names can
        easily be changed with the `rename` method of xarray Datasets/DataArrays.

        Note that ds need not contain all of these variables, and this
        function will still provide as many diagnostics as possible.

    verbose : bool, optional
        Whether to print out progress information as the function proceeds.
        Defaults to False.
    include_waves : bool, optional
        Whether to include diagnostics as a function of zonal wavenumber such
        as eddy covariances and fourier coefficients. Defaults to False.
    waves : array-like, optional
        The specific zonal wavenumbers to maintain in the output. This
        kwarg is only considered if include_waves is True.
    fftpkg : string, optional
        String that specifies how to perform the FFT on the data. Options are
        'scipy' or 'xrft'. Specifying scipy uses some operations that are memory-eager
        and leverages scipy.fft.rfft. Specifying xrft should leverage the benefits
        of xarray/dask for large datasets by using xrft.fft. Defaults to scipy.
        This kwarg is only considered if include_waves is True.

    Returns
    -------
    `xarray.Dataset`
        An xarray Dataset containing the possible zonal mean diagnostics.

    Notes
    -----
    Please see https://essd.copernicus.org/articles/10/1925/2018/ for
    a description of a different zonal mean dataset compiled for the
    SPARC Reanalysis Intercomparison Project. This function does *not*
    provide all the same diagnostics as listed in that publication.
    However, if this function is provided with all of u, v, w, and T,
    it will return all the terms necessary to compute further diagnostics
    for, e.g., zonal Eulerian and Transformed Eulerian Mean momentum budgets.

    """

    logger = logging.getLogger("create_zonal_mean_dataset")
    logger.setLevel(logging.INFO if verbose else logging.WARNING)

    if lon_coord == "":
        coords = infer_xr_coord_names(ds, required=["lon"])
        lon_coord = coords["lon"]

    valid_vars = ("u", "v", "w", "T", "Z")
    valid_covs = (("u", "v"), ("v", "T"), ("u", "w"), ("w", "T"))
    wave_vars = ("T", "Z")

    vars_available = [v for v in ds.data_vars if v in valid_vars]
    if len(vars_available) == 0:
        raise ValueError("No valid fields found in provided dataset")
    cov_pairs = [
        (v1, v2)
        for v1, v2 in valid_covs
        if v1 in vars_available and v2 in vars_available
    ]

    inter = {}
    out_coords = None
    logger.info(" *** Computing zonal means and eddies")
    for var in vars_available:
        logger.info(f"   {var}")

        zm = zonal_mean(ds[var])
        ed = ds[var] - zm

        inter[f"{var}"] = zm
        inter[f"{var}ed"] = ed
        out_coords = inter[f"{var}"].coords

    # another for-loop is not strictly necessary, but it makes the
    # logging messages a bit cleaner
    if include_waves is True:
        logger.info(" *** Computing zonal fourier transforms")
        for var in vars_available:
            logger.info(f"   fft({var})")
            fc = zonal_wave_coeffs(ds[var], waves=waves, fftpkg=fftpkg)
            if var in wave_vars:
                inter[f"{var}_k_real"] = fc.real
                inter[f"{var}_k_imag"] = fc.imag
            inter[f"{var}fc"] = fc
            out_coords = inter[f"{var}fc"].coords

    if len(cov_pairs) > 0:
        logger.info(" *** Computing eddy covariances")
        for var1, var2 in cov_pairs:
            logger.info(f"   {var1}{var2}")

            cov = zonal_mean(inter[f"{var1}ed"] * inter[f"{var2}ed"])
            inter[f"{var1}{var2}"] = cov
            if include_waves is True:
                logger.info(f"   {var1}{var2}_k")
                cov = zonal_wave_covariance(inter[f"{var1}fc"], inter[f"{var2}fc"])
                inter[f"{var1}{var2}_k"] = cov

    # remove unneeded fields
    out_vars = list(inter.keys())
    for var in out_vars:
        if "ed" in var or "fc" in var:
            inter.pop(var)

    # add attributes
    out_vars = inter.keys()
    for var in out_vars:
        inter[var].name = var
        inter[var].attrs["long_name"] = LONG_NAMES[var]
        inter[var].attrs["units"] = UNITS[var]

    out_ds = xr.Dataset(inter, coords=out_coords)
    out_ds.attrs["nlons"] = ds[lon_coord].size
    out_ds.attrs["lon0"] = ds[lon_coord].values[0]

    return out_ds



import pytest

import numpy as np

from pyzome.tem import (
    resid_vel,
    epflux_vector,
    qg_epflux_vector,
    epflux_div,
)
from pyzome.mock_data import create_dummy_geo_dataset, lat_coord, plev_coord
from pyzome.constants import EARTH_RADIUS


def test_resid_vel():
    """Test that resid_vel works correctly"""
    ds = create_dummy_geo_dataset(
        ["v", "T", "omega", "vT"],
        lons=None,
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        field_attrs={
            "v": {"units": "m/s"},
            "omega": {"units": "Pa/s"},
            "T": {"units": "K"},
            "vT": {"units": "K m s-1"},
        },
    )

    v_star, w_star = resid_vel(ds.v, ds.omega, ds.T, ds.vT)
    assert v_star.attrs["units"] == "m s-1"
    assert w_star.attrs["units"] == "Pa s-1"


def test_epflux_vector():
    ds = create_dummy_geo_dataset(
        ["u", "T", "uv", "vT", "uw"],
        lons=None,
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        field_attrs={
            "u": {"units": "m/s"},
            "T": {"units": "K"},
            "uv": {"units": "m+2 s-2"},
            "vT": {"units": "K m s-1"},
            "uw": {"units": "m Pa s-2"},
        },
    )

    F_lat, F_prs = epflux_vector(ds.u, ds.T, ds.uv, ds.vT, ds.uw)
    assert F_lat.attrs["units"] == "m+3 s-2"
    assert F_prs.attrs["units"] == "Pa m+2 s-2"


def test_qg_epflux_vector():
    ds = create_dummy_geo_dataset(
        ["T", "uv", "vT"],
        lons=None,
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        field_attrs={
            "T": {"units": "K"},
            "uv": {"units": "m+2 s-2"},
            "vT": {"units": "K m s-1"},
        },
    )

    F_lat, F_prs = qg_epflux_vector(ds.T, ds.uv, ds.vT)
    assert F_lat.attrs["units"] == "m+3 s-2"
    assert F_prs.attrs["units"] == "Pa m+2 s-2"


def test_epflux_div():
    ds = create_dummy_geo_dataset(
        ["F_lat", "F_prs"],
        lons=None,
        lats=lat_coord(2.5),
        levs=plev_coord(5, 2, units="Pa"),
        field_attrs={
            "F_lat": {"units": "m+3 s-2"},
            "F_prs": {"units": "Pa m+2 s-2"},
        },
    )

    div = epflux_div(ds.F_lat, ds.F_prs)
    assert div.attrs["units"] == "m+2 s-2"

    cos_lats = np.cos(np.deg2rad(ds.lat))
    scale = 1 / (EARTH_RADIUS * cos_lats)
    div_accel = epflux_div(ds.F_lat, ds.F_prs, accel=True)
    assert div_accel.attrs["units"] == "m s-2"
    np.testing.assert_allclose((div * scale).values, div_accel.values)

    merid_div, verti_div = epflux_div(ds.F_lat, ds.F_prs, terms=True)
    assert merid_div.attrs["units"] == div.attrs["units"]
    assert verti_div.attrs["units"] == div.attrs["units"]



import pytest

import numpy as np

from pyzome.qglogp import (
    add_logp_altitude,
    buoyancy_frequency_squared,
    merid_grad_qgpv,
    refractive_index,
    plumb_wave_activity_flux,
)
from pyzome.mock_data import (
    create_dummy_geo_dataset, 
    create_dummy_geo_field, 
    lon_coord, 
    lat_coord, 
    plev_coord
)
from pyzome.constants import SCALE_HEIGHT, PREF


def test_add_logp_altitude():
    """Test that add_logp_altitude works correctly"""
    da = create_dummy_geo_field(
        lons=lon_coord(2.5),
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        name = "T",
        attrs={"units": "K"},
    )

    da = add_logp_altitude(da)
    assert da.z.attrs["units"] == "m"
    assert da.z.attrs["long_name"] == "log-pressure altitude"
    assert da.z.attrs["note"] == "added by pyzome"
    np.testing.assert_allclose(da.z.values, -SCALE_HEIGHT*np.log(da.lev.values/PREF))


def test_buoyancy_frequency_squared():
    """Test that buoyancy_frequency_squared computation works from start to finish"""
    T = create_dummy_geo_field(
        lons=None,
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        name = "T",
        attrs={"units": "K"},
    )
    T = add_logp_altitude(T)

    N2 = buoyancy_frequency_squared(T)
    assert N2.attrs["units"] == "s-2"
    assert N2.attrs["long_name"] == "buoyancy frequency squared"


def test_merid_grad_qgpv():
    """Test that merid_grad_qgpv computation works from start to finish"""
    ds = create_dummy_geo_dataset(
        field_names = ["u", "N2"],
        lons=None,
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        field_attrs={"u": {"units": "m s-1"}, "N2": {"units": "s-2"}},
    )
    ds = add_logp_altitude(ds)

    dqdy = merid_grad_qgpv(ds.u, ds.N2)
    term1, term2, term3 = merid_grad_qgpv(ds.u, ds.N2, terms=True)
    np.testing.assert_allclose(dqdy.values, (term1 + term2 + term3).values)


def test_refractive_index():
    """Test that refractive_index computation works from start to finish"""

    ds = create_dummy_geo_dataset(
        field_names = ["u", "q_phi", "N2"],
        lons=None,
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        field_attrs={"u": {"units": "m s-1"}, "q_phi": {"units": "s-1"}, "N2": {"units": "s-2"}},
    )
    ds = add_logp_altitude(ds)

    ri2 = refractive_index(ds.u, ds.q_phi, np.abs(ds.N2), 1)
    term1, term2, term3  = refractive_index(ds.u, ds.q_phi, np.abs(ds.N2), 1, terms=True)
    np.testing.assert_allclose(ri2.values, (term1 + term2 + term3).values)

    # Test with alternate buoyancy frequency term
    ri2 = refractive_index(ds.u, ds.q_phi, np.abs(ds.N2), 1, cd_approx_term=True)
    term1, term2, term3 = refractive_index(ds.u, ds.q_phi, np.abs(ds.N2), 1, terms=True, cd_approx_term=True)
    np.testing.assert_allclose(ri2.values, (term1 + term2 + term3).values)


def test_plumb_wave_activity_flux():
    """Test that plumb_wave_activity_flux computation works from start to finish"""
    
    ds = create_dummy_geo_dataset(
        field_names = ["psi", "N2"],
        lons=lon_coord(2.5),
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        field_attrs={"psi": {"units": "m+2 s-1"}, "N2": {"units": "s-2"}},
    )
    ds = add_logp_altitude(ds)

    combos = [["x","y","z"], ["x","z"], ["y","z"], ["x"], ["y"], ["z"]]
    for components in combos:
        F = plumb_wave_activity_flux(ds.psi, ds.N2, components=components)
        assert len(F) == len(components)


def test_plumb_wave_activity_flux_no_lats():
    ds = create_dummy_geo_dataset(
        field_names = ["psi", "N2"],
        lons=lon_coord(2.5),
        lats=None,
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        field_attrs={"psi": {"units": "m+2 s-1"}, "N2": {"units": "s-2"}},
    )
    ds = add_logp_altitude(ds)

    with pytest.raises(ValueError) as e:
        F = plumb_wave_activity_flux(ds.psi, ds.N2, components=["y"])
    assert "A latitude coordinate must be available" in str(e.value)


def test_plumb_wave_activity_flux_no_lons():
    ds = create_dummy_geo_dataset(
        field_names = ["psi", "N2"],
        lons=None,
        lats=lat_coord(2.5),
        levs=plev_coord(3, left_lim_exponent=5, right_lim_exponent=2, units="Pa"),
        field_attrs={"psi": {"units": "m+2 s-1"}, "N2": {"units": "s-2"}},
    )
    ds = add_logp_altitude(ds)

    with pytest.raises(ValueError) as e:
        F = plumb_wave_activity_flux(ds.psi, ds.N2, components=["x", "z"])
    assert "A longitude coordinate must be available" in str(e.value)
